# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import urllib.error

from glance_store._drivers import filesystem
from oslo_config import cfg
from oslo_utils.fixture import uuidsentinel

from glance.async_.flows._internal_plugins import glance_download
from glance.async_.flows import api_image_import
from glance.common import exception
from glance.common.scripts import utils as script_utils
import glance.context
from glance import domain
import glance.tests.utils as test_utils

CONF = cfg.CONF


TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'


class TestGlanceDownloadTask(test_utils.BaseTestCase):

    def setUp(self):
        super(TestGlanceDownloadTask, self).setUp()

        self.config(node_staging_uri='/tmp/staging')
        self.image_repo = mock.MagicMock()
        self.image_id = mock.MagicMock()
        self.uri = mock.MagicMock()
        self.task_factory = domain.TaskFactory()
        self.context = glance.context.RequestContext(tenant=TENANT1,
                                                     auth_token='token')
        task_input = {
            "import_req": {
                'method': {
                    'name': 'glance-download',
                    'glance_image_id': uuidsentinel.remote_image,
                    'glance_region': 'RegionTwo',
                    'glance_service_interface': 'public',
                }
            }
        }
        task_ttl = CONF.task.task_time_to_live

        self.task_type = 'import'
        request_id = 'fake_request_id'
        user_id = 'fake_user'
        self.task = self.task_factory.new_task(self.task_type, TENANT1,
                                               self.image_id, user_id,
                                               request_id,
                                               task_time_to_live=task_ttl,
                                               task_input=task_input)

        self.task_id = self.task.task_id
        self.action_wrapper = api_image_import.ImportActionWrapper(
            self.image_repo, self.image_id, self.task_id)
        self.image_repo.get.return_value = mock.MagicMock(
            extra_properties={'os_glance_import_task': self.task_id})

    @mock.patch('glance.common.utils.socket.getaddrinfo')
    @mock.patch.object(filesystem.Store, 'add')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_glance_download(self, mock_gge, mock_add, mock_getaddrinfo):
        mock_getaddrinfo.return_value = [('', '', '', '', ('', 80))]
        mock_gge.return_value = 'https://other.cloud.foo/image'
        glance_download_task = glance_download._DownloadGlanceImage(
            self.context, self.task.task_id, self.task_type,
            self.action_wrapper, ['foo'],
            'RegionTwo', uuidsentinel.remote_image, 'public')
        with mock.patch('urllib.request') as mock_request:
            mock_opener = mock.MagicMock()
            mock_response = mock.MagicMock()
            mock_opener.open.return_value = mock_response
            mock_request.build_opener.return_value = mock_opener
            mock_add.return_value = ["path", 12345]
            self.assertEqual(glance_download_task.execute(12345), "path")
            mock_add.assert_called_once_with(
                self.image_id,
                mock_response, 0)
            mock_request.Request.assert_called_once_with(
                'https://other.cloud.foo/image/v2/images/%s/file' % (
                    uuidsentinel.remote_image),
                headers={'X-Auth-Token': self.context.auth_token})
            mock_request.build_opener.assert_called_once()
        mock_gge.assert_called_once_with(self.context, 'RegionTwo', 'public')

    @mock.patch('glance.common.utils.socket.getaddrinfo')
    @mock.patch.object(filesystem.Store, 'add')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_glance_download_failed(self, mock_gge, mock_add,
                                    mock_getaddrinfo):
        mock_getaddrinfo.return_value = [('', '', '', '', ('', 80))]
        mock_gge.return_value = 'https://other.cloud.foo/image'
        glance_download_task = glance_download._DownloadGlanceImage(
            self.context, self.task.task_id, self.task_type,
            self.action_wrapper, ['foo'],
            'RegionTwo', uuidsentinel.remote_image, 'public')
        with mock.patch('urllib.request') as mock_request:
            mock_opener = mock.MagicMock()
            mock_opener.open.side_effect = urllib.error.HTTPError(
                '/file', 400, 'Test Fail', {}, None)
            mock_request.build_opener.return_value = mock_opener
            self.assertRaises(urllib.error.HTTPError,
                              glance_download_task.execute,
                              12345)
            mock_add.assert_not_called()
            mock_request.Request.assert_called_once_with(
                'https://other.cloud.foo/image/v2/images/%s/file' % (
                    uuidsentinel.remote_image),
                headers={'X-Auth-Token': self.context.auth_token})
        mock_gge.assert_called_once_with(self.context, 'RegionTwo', 'public')

    @mock.patch('urllib.request')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_glance_download_no_glance_endpoint(self, mock_gge, mock_request):
        mock_gge.side_effect = glance.common.exception.GlanceEndpointNotFound(
            region='RegionTwo',
            interface='public')
        glance_download_task = glance_download._DownloadGlanceImage(
            self.context, self.task.task_id, self.task_type,
            self.action_wrapper, ['foo'],
            'RegionTwo', uuidsentinel.remote_image, 'public')
        self.assertRaises(glance.common.exception.GlanceEndpointNotFound,
                          glance_download_task.execute, 12345)
        mock_request.assert_not_called()

    @mock.patch('glance.common.utils.socket.getaddrinfo')
    @mock.patch.object(filesystem.Store, 'add')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_glance_download_size_mismatch(self, mock_gge, mock_add,
                                           mock_getaddrinfo):
        mock_getaddrinfo.return_value = [('', '', '', '', ('', 80))]
        mock_gge.return_value = 'https://other.cloud.foo/image'
        glance_download_task = glance_download._DownloadGlanceImage(
            self.context, self.task.task_id, self.task_type,
            self.action_wrapper, ['foo'],
            'RegionTwo', uuidsentinel.remote_image, 'public')
        with mock.patch('urllib.request') as mock_request:
            mock_opener = mock.MagicMock()
            mock_response = mock.MagicMock()
            mock_opener.open.return_value = mock_response
            mock_request.build_opener.return_value = mock_opener
            mock_add.return_value = ["path", 1]
            self.assertRaises(glance.common.exception.ImportTaskError,
                              glance_download_task.execute, 12345)
            mock_add.assert_called_once_with(
                self.image_id,
                mock_response, 0)
            mock_request.Request.assert_called_once_with(
                'https://other.cloud.foo/image/v2/images/%s/file' % (
                    uuidsentinel.remote_image),
                headers={'X-Auth-Token': self.context.auth_token})
        mock_gge.assert_called_once_with(self.context, 'RegionTwo', 'public')

    @mock.patch('urllib.request')
    @mock.patch('glance.common.utils.validate_import_uri')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_glance_download_wrong_download_url(self, mock_gge, mock_validate,
                                                mock_request):
        mock_validate.return_value = False
        mock_gge.return_value = 'https://other.cloud.foo/image'
        glance_download_task = glance_download._DownloadGlanceImage(
            self.context, self.task.task_id, self.task_type,
            self.action_wrapper, ['foo'],
            'RegionTwo', uuidsentinel.remote_image, 'public')
        self.assertRaises(glance.common.exception.ImportTaskError,
                          glance_download_task.execute, 12345)
        mock_request.assert_not_called()
        mock_validate.assert_called_once_with(
            'https://other.cloud.foo/image/v2/images/%s/file' % (
                uuidsentinel.remote_image))

    @mock.patch('glance.common.utils.socket.getaddrinfo')
    @mock.patch('urllib.request')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_glance_download_redirect_validation(self, mock_gge,
                                                 mock_request,
                                                 mock_getaddrinfo):
        """Test redirect destinations are validated during image download."""
        mock_getaddrinfo.return_value = [('', '', '', '', ('', 80))]
        mock_gge.return_value = 'https://other.cloud.foo/image'
        glance_download_task = glance_download._DownloadGlanceImage(
            self.context, self.task.task_id, self.task_type,
            self.action_wrapper, ['foo'],
            'RegionTwo', uuidsentinel.remote_image, 'public')
        mock_opener = mock.MagicMock()
        # Simulate redirect to disallowed URL
        mock_opener.open.side_effect = exception.ImportTaskError(
            "Redirect to disallowed URL: http://127.0.0.1:5000/")
        mock_request.build_opener.return_value = mock_opener
        self.assertRaises(exception.ImportTaskError,
                          glance_download_task.execute, 12345)
        # Verify SafeRedirectHandler is used
        mock_request.build_opener.assert_called_once()
        # Verify the handler passed is SafeRedirectHandler
        call_args = mock_request.build_opener.call_args
        # Check if SafeRedirectHandler class or instance is in args
        found_handler = (
            any(isinstance(arg, script_utils.SafeRedirectHandler)
                for arg in call_args.args) or
            script_utils.SafeRedirectHandler in call_args.args)
        self.assertTrue(
            found_handler,
            "SafeRedirectHandler should be used for redirect validation")

    @mock.patch('glance.common.utils.socket.getaddrinfo')
    @mock.patch.object(filesystem.Store, 'add')
    @mock.patch('urllib.request')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_glance_download_uses_safe_redirect_handler(
            self, mock_gge, mock_request, mock_add, mock_getaddrinfo):
        """Test that SafeRedirectHandler is used and allows valid execution."""
        mock_getaddrinfo.return_value = [('', '', '', '', ('', 80))]
        mock_gge.return_value = 'https://other.cloud.foo/image'
        glance_download_task = glance_download._DownloadGlanceImage(
            self.context, self.task.task_id, self.task_type,
            self.action_wrapper, ['foo'],
            'RegionTwo', uuidsentinel.remote_image, 'public')
        mock_opener = mock.MagicMock()
        mock_response = mock.MagicMock()
        mock_opener.open.return_value = mock_response
        mock_request.build_opener.return_value = mock_opener
        mock_add.return_value = ["path", 12345]
        result = glance_download_task.execute(12345)
        # Verify build_opener was called with SafeRedirectHandler
        mock_request.build_opener.assert_called_once()
        # Verify SafeRedirectHandler was passed
        call_args = mock_request.build_opener.call_args
        # Check if SafeRedirectHandler class or instance is in args
        found_handler = (
            any(isinstance(arg, script_utils.SafeRedirectHandler)
                for arg in call_args.args) or
            script_utils.SafeRedirectHandler in call_args.args)
        self.assertTrue(
            found_handler,
            "SafeRedirectHandler should be passed to build_opener")
        # Verify execution succeeded (handler allows valid execution)
        self.assertEqual("path", result)



# Copyright 2011-2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
import webob.exc

from glance.api import policy
from glance.common import wsgi
import glance.context
from glance.i18n import _, _LW


context_opts = [
    cfg.BoolOpt('allow_anonymous_access', default=False,
                help=_("""
Allow limited access to unauthenticated users.

Assign a boolean to determine API access for unauthenticated
users. When set to False, the API cannot be accessed by
unauthenticated users. When set to True, unauthenticated users can
access the API with read-only privileges. This however only applies
when using ContextMiddleware.

Possible values:
    * True
    * False

Related options:
    * None

""")),

    cfg.IntOpt('max_request_id_length', default=64, min=0,
               help=_("""
Limit the request ID length.

Provide  an integer value to limit the length of the request ID to
the specified length. The default value is 64. Users can change this
to any ineteger value between 0 and 16384 however keeping in mind that
a larger value may flood the logs.

Possible values:
    * Integer value between 0 and 16384

Related options:
    * None

""")),
]

CONF = cfg.CONF
CONF.register_opts(context_opts)

LOG = logging.getLogger(__name__)


class BaseContextMiddleware(wsgi.Middleware):
    def process_response(self, resp):
        try:
            request_id = resp.request.context.request_id
        except AttributeError:
            LOG.warning(_LW('Unable to retrieve request id from context'))
        else:
            # For python 3 compatibility need to use bytes type
            prefix = b'req-' if isinstance(request_id, bytes) else 'req-'

            if not request_id.startswith(prefix):
                request_id = prefix + request_id

            resp.headers['x-openstack-request-id'] = request_id

        return resp


class ContextMiddleware(BaseContextMiddleware):
    def __init__(self, app):
        self.policy_enforcer = policy.Enforcer()
        super(ContextMiddleware, self).__init__(app)

    def process_request(self, req):
        """Convert authentication information into a request context

        Generate a glance.context.RequestContext object from the available
        authentication headers and store on the 'context' attribute
        of the req object.

        :param req: wsgi request object that will be given the context object
        :raises webob.exc.HTTPUnauthorized: when value of the
                                            X-Identity-Status  header is not
                                            'Confirmed' and anonymous access
                                            is disallowed
        """
        if req.headers.get('X-Identity-Status') == 'Confirmed':
            req.context = self._get_authenticated_context(req)
        elif CONF.allow_anonymous_access:
            req.context = self._get_anonymous_context()
        else:
            raise webob.exc.HTTPUnauthorized()

    def _get_anonymous_context(self):
        kwargs = {
            'user': None,
            'tenant': None,
            'roles': [],
            'is_admin': False,
            'read_only': True,
            'policy_enforcer': self.policy_enforcer,
        }
        return glance.context.RequestContext(**kwargs)

    def _get_authenticated_context(self, req):
        service_catalog = None
        if req.headers.get('X-Service-Catalog') is not None:
            try:
                catalog_header = req.headers.get('X-Service-Catalog')
                service_catalog = jsonutils.loads(catalog_header)
            except ValueError:
                raise webob.exc.HTTPInternalServerError(
                    _('Invalid service catalog json.'))

        request_id = req.headers.get('X-Openstack-Request-ID')
        if request_id and (0 < CONF.max_request_id_length <
                           len(request_id)):
            msg = (_('x-openstack-request-id is too long, max size %s') %
                   CONF.max_request_id_length)
            return webob.exc.HTTPRequestHeaderFieldsTooLarge(comment=msg)

        kwargs = {
            'service_catalog': service_catalog,
            'policy_enforcer': self.policy_enforcer,
            'request_id': request_id,
        }

        ctxt = glance.context.RequestContext.from_environ(req.environ,
                                                          **kwargs)

        # FIXME(jamielennox): glance has traditionally lowercased its roles.
        # This was related to bug #1010519 where at least the admin role was
        # case insensitive. This seems to no longer be the case and should be
        # fixed.
        ctxt.roles = [r.lower() for r in ctxt.roles]

        return ctxt


class UnauthenticatedContextMiddleware(BaseContextMiddleware):
    def process_request(self, req):
        """Create a context without an authorized user."""
        kwargs = {
            'user': None,
            'tenant': None,
            'roles': [],
            'is_admin': True,
        }

        req.context = glance.context.RequestContext(**kwargs)



# Copyright (C) 2021 RedHat Inc
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add image_id, request_id, user columns to tasks table"

Revision ID: wallaby_expand01
Revises: ussuri_expand01
Create Date: 2021-02-04 11:55:16.657499

"""

from alembic import op
from sqlalchemy import String, Column

# revision identifiers, used by Alembic.
revision = 'wallaby_expand01'
down_revision = 'ussuri_expand01'
branch_labels = None
depends_on = None


def upgrade():
    image_id_col = Column('image_id', String(length=36), nullable=True)
    request_id_col = Column('request_id', String(length=64), nullable=True)
    user_col = Column('user_id', String(length=64), nullable=True)
    op.add_column('tasks', image_id_col)
    op.add_column('tasks', request_id_col)
    op.add_column('tasks', user_col)
    op.create_index('ix_tasks_image_id', 'tasks', ['image_id'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import utils as db_utils

from glance.db.sqlalchemy.alembic_migrations import data_migrations
from glance.tests.functional.db import test_migrations
import glance.tests.utils as test_utils


class TestOcataContract01Mixin(test_migrations.AlembicMigrationsMixin):

    def _get_revisions(self, config):
        return test_migrations.AlembicMigrationsMixin._get_revisions(
            self, config, head='ocata_contract01')

    def _pre_upgrade_ocata_contract01(self, engine):
        images = db_utils.get_table(engine, 'images')
        now = datetime.datetime.now()
        self.assertIn('is_public', images.c)
        self.assertIn('visibility', images.c)
        self.assertTrue(images.c.is_public.nullable)
        self.assertTrue(images.c.visibility.nullable)

        # inserting a public image record
        public_temp = dict(deleted=False,
                           created_at=now,
                           status='active',
                           is_public=True,
                           min_disk=0,
                           min_ram=0,
                           id='public_id_before_expand')
        with engine.connect() as conn, conn.begin():
            conn.execute(images.insert().values(public_temp))

        # inserting a private image record
        shared_temp = dict(deleted=False,
                           created_at=now,
                           status='active',
                           is_public=False,
                           min_disk=0,
                           min_ram=0,
                           id='private_id_before_expand')
        with engine.connect() as conn, conn.begin():
            conn.execute(images.insert().values(shared_temp))

        data_migrations.migrate(engine=engine, release='ocata')

    def _check_ocata_contract01(self, engine, data):
        # check that after contract 'is_public' column is dropped
        images = db_utils.get_table(engine, 'images')
        self.assertNotIn('is_public', images.c)
        self.assertIn('visibility', images.c)


class TestOcataContract01MySQL(
    TestOcataContract01Mixin,
    test_fixtures.OpportunisticDBTestMixin,
    test_utils.BaseTestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client as http
import io
from unittest import mock
import uuid

from cursive import exception as cursive_exception
import glance_store
from glance_store._drivers import filesystem
from oslo_config import cfg
import webob

import glance.api.policy
import glance.api.v2.image_data
from glance.common import exception
from glance.common import wsgi
from glance.tests.unit import base
import glance.tests.unit.utils as unit_test_utils
import glance.tests.utils as test_utils

CONF = cfg.CONF
CONF.import_opt('public_endpoint', 'glance.api.versions')


class Raise(object):

    def __init__(self, exc):
        self.exc = exc

    def __call__(self, *args, **kwargs):
        raise self.exc


class FakeImage(object):

    def __init__(self, image_id=None, data=None, checksum=None, size=0,
                 virtual_size=0, locations=None, container_format='bear',
                 disk_format='rawr', status=None, owner=None):
        self.image_id = image_id
        self.data = data
        self.checksum = checksum
        self.size = size
        self.virtual_size = virtual_size
        self.locations = locations
        self.container_format = container_format
        self.disk_format = disk_format
        self.owner = owner
        self._status = status
        self.extra_properties = {}

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        if isinstance(self._status, BaseException):
            raise self._status
        else:
            self._status = value

    def get_data(self, offset=0, chunk_size=None):
        if chunk_size:
            return self.data[offset:offset + chunk_size]
        return self.data[offset:]

    def set_data(self, data, size=None, backend=None, set_active=True):
        self.data = ''.join(data)
        self.size = size
        self.status = 'modified-by-fake'


class FakeImageRepo(object):

    def __init__(self, result=None):
        self.result = result

    def get(self, image_id):
        if isinstance(self.result, BaseException):
            raise self.result
        else:
            return self.result

    def save(self, image, from_state=None):
        self.saved_image = image


class FakeGateway(object):

    def __init__(self, db=None, store=None, notifier=None,
                 policy=None, repo=None):
        self.db = db
        self.store = store
        self.notifier = notifier
        self.policy = policy
        self.repo = repo

    def get_repo(self, context):
        return self.repo


class TestImagesController(base.StoreClearingUnitTest):

    def setUp(self):
        super(TestImagesController, self).setUp()

        self.config(debug=True)
        self.image_repo = FakeImageRepo()
        db = unit_test_utils.FakeDB()
        policy = unit_test_utils.FakePolicyEnforcer()
        notifier = unit_test_utils.FakeNotifier()
        store = unit_test_utils.FakeStoreAPI()
        self.controller = glance.api.v2.image_data.ImageDataController()
        self.controller.gateway = FakeGateway(db, store, notifier, policy,
                                              self.image_repo)
        # FIXME(abhishekk): Everything is fake in this test, so mocked the
        # image mutable_check, Later we need to fix these tests to use
        # some realistic data
        patcher = mock.patch('glance.api.v2.policy.check_is_image_mutable')
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_download(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('abcd',
                          locations=[{'url': 'http://example.com/image',
                                      'metadata': {}, 'status': 'active'}])
        self.image_repo.result = image
        image = self.controller.download(request, unit_test_utils.UUID1)
        self.assertEqual('abcd', image.image_id)

    def test_download_deactivated(self):
        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd',
                          status='deactivated',
                          locations=[{'url': 'http://example.com/image',
                                      'metadata': {}, 'status': 'active'}])
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPForbidden, self.controller.download,
                          request, str(uuid.uuid4()))

    def test_download_no_location(self):
        # NOTE(mclaren): NoContent will be raised by the ResponseSerializer
        # That's tested below.
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        self.image_repo.result = FakeImage('abcd')
        image = self.controller.download(request, unit_test_utils.UUID2)
        self.assertEqual('abcd', image.image_id)

    def test_download_non_existent_image(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.NotFound()
        self.assertRaises(webob.exc.HTTPNotFound, self.controller.download,
                          request, str(uuid.uuid4()))

    def test_download_forbidden(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.Forbidden()
        self.assertRaises(webob.exc.HTTPForbidden, self.controller.download,
                          request, str(uuid.uuid4()))

    def test_download_ok_when_get_image_location_forbidden(self):
        class ImageLocations(object):

            def __len__(self):
                raise exception.Forbidden()

        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('abcd')
        self.image_repo.result = image
        image.locations = ImageLocations()
        image = self.controller.download(request, unit_test_utils.UUID1)
        self.assertEqual('abcd', image.image_id)

    def test_upload(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('abcd', owner='tenant1')
        self.image_repo.result = image
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        self.assertEqual('YYYY', image.data)
        self.assertEqual(4, image.size)

    def test_upload_not_allowed_by_policy(self):
        request = unit_test_utils.get_fake_request()
        with mock.patch.object(self.controller.policy, 'enforce') as mock_enf:
            mock_enf.side_effect = webob.exc.HTTPForbidden()
            exc = self.assertRaises(webob.exc.HTTPNotFound,
                                    self.controller.upload, request,
                                    unit_test_utils.UUID1, 'YYYY', 4)
            self.assertTrue(mock_enf.called)
        # Make sure we did not leak details of the original Forbidden
        # error into the NotFound returned to the client.
        self.assertEqual('The resource could not be found.', str(exc))

        # Now reject the upload_image call, but allow get_image to ensure that
        # we properly see a Forbidden result.
        with mock.patch.object(self.controller.policy, 'enforce') as mock_enf:
            mock_enf.side_effect = [webob.exc.HTTPForbidden(),
                                    lambda *a: None]
            exc = self.assertRaises(webob.exc.HTTPForbidden,
                                    self.controller.upload, request,
                                    unit_test_utils.UUID1, 'YYYY', 4)
            self.assertTrue(mock_enf.called)

    def test_upload_status(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('abcd')
        self.image_repo.result = image
        insurance = {'called': False}

        def read_data():
            insurance['called'] = True
            self.assertEqual('saving', self.image_repo.saved_image.status)
            yield 'YYYY'

        self.controller.upload(request, unit_test_utils.UUID2,
                               read_data(), None)
        self.assertTrue(insurance['called'])
        self.assertEqual('modified-by-fake',
                         self.image_repo.saved_image.status)

    def test_upload_no_size(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('abcd')
        self.image_repo.result = image
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', None)
        self.assertEqual('YYYY', image.data)
        self.assertIsNone(image.size)

    @mock.patch.object(glance.api.policy.Enforcer, 'enforce')
    def test_upload_image_forbidden(self, mock_enforce):
        request = unit_test_utils.get_fake_request()
        image = FakeImage('abcd', owner='tenant1')
        self.image_repo.result = image
        mock_enforce.side_effect = [exception.Forbidden, lambda *a: None]
        self.assertRaises(webob.exc.HTTPForbidden, self.controller.upload,
                          request, unit_test_utils.UUID2, 'YYYY', 4)
        expected_call = [
            mock.call(mock.ANY, 'upload_image', mock.ANY),
            mock.call(mock.ANY, 'get_image', mock.ANY)
        ]
        mock_enforce.assert_has_calls(expected_call)

    def test_upload_invalid(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('abcd')
        image.status = ValueError()
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYY', 4)

    def test_upload_with_expired_token(self):
        def side_effect(image, from_state=None):
            if from_state == 'saving':
                raise exception.NotAuthenticated()

        mocked_save = mock.Mock(side_effect=side_effect)
        mocked_delete = mock.Mock()
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('abcd')
        image.delete = mocked_delete
        self.image_repo.result = image
        self.image_repo.save = mocked_save
        self.assertRaises(webob.exc.HTTPUnauthorized, self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYY', 4)
        self.assertEqual(3, mocked_save.call_count)
        mocked_delete.assert_called_once_with()

    @mock.patch('glance.common.trust_auth.TokenRefresher')
    def test_upload_with_token_refresh(self, mock_refresher):
        mock_refresher.return_value = mock.MagicMock()
        mocked_save = mock.Mock()
        mocked_save.side_effect = [lambda *a: None,
                                   exception.NotAuthenticated(),
                                   lambda *a: None]
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        request.environ['keystone.token_info'] = {
            'token': {
                'roles': [{'name': 'member'}]
            }
        }
        image = FakeImage('abcd', owner='tenant1')
        self.image_repo.result = image
        self.image_repo.save = mocked_save
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        self.assertEqual('YYYY', image.data)
        self.assertEqual(4, image.size)
        self.assertEqual(3, mocked_save.call_count)

    def test_upload_non_existent_image_during_save_initiates_deletion(self):
        def fake_save_not_found(self, from_state=None):
            raise exception.ImageNotFound()

        def fake_save_conflict(self, from_state=None):
            raise exception.Conflict()

        for fun in [fake_save_not_found, fake_save_conflict]:
            request = unit_test_utils.get_fake_request(
                roles=['admin', 'member'])
            image = FakeImage('abcd', locations=['http://example.com/image'])
            self.image_repo.result = image
            self.image_repo.save = fun
            image.delete = mock.Mock()
            self.assertRaises(webob.exc.HTTPGone, self.controller.upload,
                              request, str(uuid.uuid4()), 'ABC', 3)
            self.assertTrue(image.delete.called)

    def test_upload_non_existent_image_raises_image_not_found_exception(self):
        def fake_save(self, from_state=None):
            raise exception.ImageNotFound()

        def fake_delete():
            raise exception.ImageNotFound()

        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('abcd', locations=['http://example.com/image'])
        self.image_repo.result = image
        self.image_repo.save = fake_save
        image.delete = fake_delete
        self.assertRaises(webob.exc.HTTPGone, self.controller.upload,
                          request, str(uuid.uuid4()), 'ABC', 3)

    def test_upload_non_existent_image_raises_store_not_found_exception(self):
        def fake_save(self, from_state=None):
            raise glance_store.NotFound()

        def fake_delete():
            raise exception.ImageNotFound()

        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('abcd', locations=['http://example.com/image'])
        self.image_repo.result = image
        self.image_repo.save = fake_save
        image.delete = fake_delete
        self.assertRaises(webob.exc.HTTPGone, self.controller.upload,
                          request, str(uuid.uuid4()), 'ABC', 3)

    def test_upload_non_existent_image_before_save(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.NotFound()
        self.assertRaises(webob.exc.HTTPNotFound, self.controller.upload,
                          request, str(uuid.uuid4()), 'ABC', 3)

    def test_upload_data_exists(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage()
        exc = exception.InvalidImageStatusTransition(cur_status='active',
                                                     new_status='queued')
        image.set_data = Raise(exc)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPConflict, self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYY', 4)

    def test_upload_storage_full(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage()
        image.set_data = Raise(glance_store.StorageFull)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                          self.controller.upload,
                          request, unit_test_utils.UUID2, 'YYYYYYY', 7)

    def test_upload_signature_verification_fails(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage()
        image.set_data = Raise(cursive_exception.SignatureVerificationError)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYY', 4)
        self.assertEqual('queued', self.image_repo.saved_image.status)

    def test_image_size_limit_exceeded(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage()
        image.set_data = Raise(exception.ImageSizeLimitExceeded)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                          self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYYYYY', 7)

    def test_upload_storage_quota_full(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.StorageQuotaFull("message")
        self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                          self.controller.upload,
                          request, unit_test_utils.UUID1, 'YYYYYYY', 7)

    def test_upload_storage_forbidden(self):
        request = unit_test_utils.get_fake_request(
            user=unit_test_utils.USER2, roles=['admin', 'member'])
        image = FakeImage()
        image.set_data = Raise(exception.Forbidden)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPForbidden, self.controller.upload,
                          request, unit_test_utils.UUID2, 'YY', 2)

    def test_upload_storage_internal_error(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.ServerError()
        self.assertRaises(exception.ServerError,
                          self.controller.upload,
                          request, unit_test_utils.UUID1, 'ABC', 3)

    def test_upload_storage_write_denied(self):
        request = unit_test_utils.get_fake_request(
            user=unit_test_utils.USER3, roles=['admin', 'member'])
        image = FakeImage()
        image.set_data = Raise(glance_store.StorageWriteDenied)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPServiceUnavailable,
                          self.controller.upload,
                          request, unit_test_utils.UUID2, 'YY', 2)

    def test_upload_storage_store_disabled(self):
        """Test that uploading an image file raises StoreDisabled exception"""
        request = unit_test_utils.get_fake_request(
            user=unit_test_utils.USER3, roles=['admin', 'member'])
        image = FakeImage()
        image.set_data = Raise(glance_store.StoreAddDisabled)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPGone,
                          self.controller.upload,
                          request, unit_test_utils.UUID2, 'YY', 2)

    def _test_upload_download_prepare_notification(self):
        request = unit_test_utils.get_fake_request()
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        output = self.controller.download(request, unit_test_utils.UUID2)
        output_log = self.notifier.get_logs()
        prepare_payload = output['meta'].copy()
        prepare_payload['checksum'] = None
        prepare_payload['size'] = None
        prepare_payload['virtual_size'] = None
        prepare_payload['location'] = None
        prepare_payload['status'] = 'queued'
        del prepare_payload['updated_at']
        prepare_log = {
            'notification_type': "INFO",
            'event_type': "image.prepare",
            'payload': prepare_payload,
        }
        self.assertEqual(3, len(output_log))
        prepare_updated_at = output_log[0]['payload']['updated_at']
        del output_log[0]['payload']['updated_at']
        self.assertLessEqual(prepare_updated_at, output['meta']['updated_at'])
        self.assertEqual(prepare_log, output_log[0])

    def _test_upload_download_upload_notification(self):
        request = unit_test_utils.get_fake_request()
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        output = self.controller.download(request, unit_test_utils.UUID2)
        output_log = self.notifier.get_logs()
        upload_payload = output['meta'].copy()
        upload_log = {
            'notification_type': "INFO",
            'event_type': "image.upload",
            'payload': upload_payload,
        }
        self.assertEqual(3, len(output_log))
        self.assertEqual(upload_log, output_log[1])

    def _test_upload_download_activate_notification(self):
        request = unit_test_utils.get_fake_request()
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        output = self.controller.download(request, unit_test_utils.UUID2)
        output_log = self.notifier.get_logs()
        activate_payload = output['meta'].copy()
        activate_log = {
            'notification_type': "INFO",
            'event_type': "image.activate",
            'payload': activate_payload,
        }
        self.assertEqual(3, len(output_log))
        self.assertEqual(activate_log, output_log[2])

    def test_restore_image_when_upload_failed(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('fake')
        image.set_data = Raise(glance_store.StorageWriteDenied)
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPServiceUnavailable,
                          self.controller.upload,
                          request, unit_test_utils.UUID2, 'ZZZ', 3)
        self.assertEqual('queued', self.image_repo.saved_image.status)

    @mock.patch.object(filesystem.Store, 'add')
    def test_restore_image_when_staging_failed(self, mock_store_add):
        mock_store_add.side_effect = glance_store.StorageWriteDenied()
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image_id = str(uuid.uuid4())
        image = FakeImage('fake')
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPServiceUnavailable,
                          self.controller.stage,
                          request, image_id, 'YYYYYYY', 7)
        self.assertEqual('queued', self.image_repo.saved_image.status)

    def test_stage(self):
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage(image_id=image_id)
        self.image_repo.result = image
        with mock.patch.object(filesystem.Store, 'add') as mock_add:
            mock_add.return_value = ('foo://bar', 4, 'ident', {})
            self.controller.stage(request, image_id, 'YYYY', 4)
        self.assertEqual('uploading', image.status)
        self.assertEqual(4, image.size)

    def test_image_already_on_staging(self):
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage(image_id=image_id)
        self.image_repo.result = image
        with mock.patch.object(filesystem.Store, 'add') as mock_store_add:
            mock_store_add.return_value = ('foo://bar', 4, 'ident', {})
            self.controller.stage(request, image_id, 'YYYY', 4)
            self.assertEqual('uploading', image.status)
            mock_store_add.side_effect = glance_store.Duplicate()
            self.assertEqual(4, image.size)
            self.assertRaises(webob.exc.HTTPConflict, self.controller.stage,
                              request, image_id, 'YYYY', 4)

    @mock.patch.object(glance_store.driver.Store, 'configure')
    def test_image_stage_raises_bad_store_uri(self, mock_store_configure):
        mock_store_configure.side_effect = AttributeError()
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        self.assertRaises(exception.BadStoreUri, self.controller.stage,
                          request, image_id, 'YYYY', 4)

    @mock.patch.object(filesystem.Store, 'add')
    def test_image_stage_raises_storage_full(self, mock_store_add):
        mock_store_add.side_effect = glance_store.StorageFull()
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage(image_id=image_id)
        self.image_repo.result = image
        with mock.patch.object(self.controller, "_unstage"):
            self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                              self.controller.stage,
                              request, image_id, 'YYYYYYY', 7)

    @mock.patch.object(filesystem.Store, 'add')
    def test_image_stage_raises_storage_quota_full(self, mock_store_add):
        mock_store_add.side_effect = exception.StorageQuotaFull("message")
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage(image_id=image_id)
        self.image_repo.result = image
        with mock.patch.object(self.controller, "_unstage"):
            self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                              self.controller.stage,
                              request, image_id, 'YYYYYYY', 7)

    @mock.patch.object(filesystem.Store, 'add')
    def test_image_stage_raises_storage_write_denied(self, mock_store_add):
        mock_store_add.side_effect = glance_store.StorageWriteDenied()
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage(image_id=image_id)
        self.image_repo.result = image
        with mock.patch.object(self.controller, "_unstage"):
            self.assertRaises(webob.exc.HTTPServiceUnavailable,
                              self.controller.stage,
                              request, image_id, 'YYYYYYY', 7)

    def test_image_stage_raises_internal_error(self):
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.ServerError()
        self.assertRaises(exception.ServerError,
                          self.controller.stage,
                          request, image_id, 'YYYYYYY', 7)

    def test_image_stage_non_existent_image(self):
        request = unit_test_utils.get_fake_request()
        self.image_repo.result = exception.NotFound()
        self.assertRaises(webob.exc.HTTPNotFound, self.controller.stage,
                          request, str(uuid.uuid4()), 'ABC', 3)

    @mock.patch.object(filesystem.Store, 'add')
    def test_image_stage_raises_image_size_exceeded(self, mock_store_add):
        mock_store_add.side_effect = exception.ImageSizeLimitExceeded()
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage(image_id=image_id)
        self.image_repo.result = image
        with mock.patch.object(self.controller, "_unstage"):
            self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                              self.controller.stage,
                              request, image_id, 'YYYYYYY', 7)

    @mock.patch.object(filesystem.Store, 'add')
    def test_image_stage_invalid_image_transition(self, mock_store_add):
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage(image_id=image_id)
        self.image_repo.result = image
        with mock.patch.object(filesystem.Store, 'add') as mock_add:
            mock_add.return_value = ('foo://bar', 4, 'ident', {})
            self.controller.stage(request, image_id, 'YYYY', 4)
        self.assertEqual('uploading', image.status)
        self.assertEqual(4, image.size)
        # try staging again
        mock_store_add.side_effect = exception.InvalidImageStatusTransition(
            cur_status='uploading', new_status='uploading')
        self.assertRaises(webob.exc.HTTPConflict, self.controller.stage,
                          request, image_id, 'YYYY', 4)

    def _test_image_stage_records_host(self, expected_url):
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage(image_id=image_id)
        self.image_repo.result = image
        with mock.patch.object(filesystem.Store, 'add') as mock_add:
            mock_add.return_value = ('foo://bar', 4, 'ident', {})
            self.controller.stage(request, image_id, 'YYYY', 4)
        if expected_url is None:
            self.assertNotIn('os_glance_stage_host', image.extra_properties)
        else:
            self.assertEqual(expected_url,
                             image.extra_properties['os_glance_stage_host'])

    def test_image_stage_records_host_unset(self):
        # Make sure we do not set a null staging host, if we are not configured
        # to support worker-to-worker communication.
        self._test_image_stage_records_host(None)

    def test_image_stage_records_host_public_endpoint(self):
        # Make sure we fall back to public_endpoint
        self.config(public_endpoint='http://lb.example.com')
        self._test_image_stage_records_host('http://lb.example.com')

    def test_image_stage_records_host_self_url(self):
        # Make sure worker_self_reference_url takes precedence
        self.config(worker_self_reference_url='http://worker1.example.com')
        self._test_image_stage_records_host('http://worker1.example.com')

    def test_image_stage_fail_does_not_set_host(self):
        # Make sure that if the store.add() fails, we do not claim to have the
        # image staged.
        self.config(public_endpoint='http://worker1.example.com')
        image_id = str(uuid.uuid4())
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage(image_id=image_id)
        self.image_repo.result = image
        exc_cls = glance_store.exceptions.StorageFull
        with mock.patch.object(filesystem.Store, 'add', side_effect=exc_cls):
            self.assertRaises(webob.exc.HTTPRequestEntityTooLarge,
                              self.controller.stage,
                              request, image_id, 'YYYY', 4)
        self.assertNotIn('os_glance_stage_host', image.extra_properties)


class TestImageDataDeserializer(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageDataDeserializer, self).setUp()
        self.deserializer = glance.api.v2.image_data.RequestDeserializer()

    def test_upload(self):
        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/octet-stream'
        request.body = b'YYY'
        request.headers['Content-Length'] = 3
        output = self.deserializer.upload(request)
        data = output.pop('data')
        self.assertEqual(b'YYY', data.read())
        expected = {'size': 3}
        self.assertEqual(expected, output)

    def test_upload_chunked(self):
        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/octet-stream'
        # If we use body_file, webob assumes we want to do a chunked upload,
        # ignoring the Content-Length header
        request.body_file = io.StringIO('YYY')
        output = self.deserializer.upload(request)
        data = output.pop('data')
        self.assertEqual('YYY', data.read())
        expected = {'size': None}
        self.assertEqual(expected, output)

    def test_upload_chunked_with_content_length(self):
        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/octet-stream'
        request.body_file = io.BytesIO(b'YYY')
        # The deserializer shouldn't care if the Content-Length is
        # set when the user is attempting to send chunked data.
        request.headers['Content-Length'] = 3
        output = self.deserializer.upload(request)
        data = output.pop('data')
        self.assertEqual(b'YYY', data.read())
        expected = {'size': 3}
        self.assertEqual(expected, output)

    def test_upload_with_incorrect_content_length(self):
        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/octet-stream'
        # The deserializer shouldn't care if the Content-Length and
        # actual request body length differ. That job is left up
        # to the controller
        request.body = b'YYY'
        request.headers['Content-Length'] = 4
        output = self.deserializer.upload(request)
        data = output.pop('data')
        self.assertEqual(b'YYY', data.read())
        expected = {'size': 4}
        self.assertEqual(expected, output)

    def test_upload_wrong_content_type(self):
        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/json'
        request.body = b'YYYYY'
        self.assertRaises(webob.exc.HTTPUnsupportedMediaType,
                          self.deserializer.upload, request)

        request = unit_test_utils.get_fake_request()
        request.headers['Content-Type'] = 'application/octet-st'
        request.body = b'YYYYY'
        self.assertRaises(webob.exc.HTTPUnsupportedMediaType,
                          self.deserializer.upload, request)

    def test_stage(self):
        req = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        req.headers['Content-Type'] = 'application/octet-stream'
        req.headers['Content-Length'] = 4
        req.body_file = io.BytesIO(b'YYYY')
        output = self.deserializer.stage(req)
        data = output.pop('data')
        self.assertEqual(b'YYYY', data.read())

    def test_stage_without_glance_direct(self):
        self.config(enabled_import_methods=['web-download'])
        req = unit_test_utils.get_fake_request()
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.deserializer.stage,
                          req)

    def test_stage_raises_invalid_content_type(self):
        # TODO(abhishekk): change this when import methods are
        # listed in the config file
        req = unit_test_utils.get_fake_request()
        req.headers['Content-Type'] = 'application/json'
        self.assertRaises(webob.exc.HTTPUnsupportedMediaType,
                          self.deserializer.stage,
                          req)


class TestImageDataSerializer(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageDataSerializer, self).setUp()
        self.serializer = glance.api.v2.image_data.ResponseSerializer()

    def test_download(self):
        request = wsgi.Request.blank('/')
        request.environ = {}
        response = webob.Response()
        response.request = request
        image = FakeImage(size=3, data=[b'Z', b'Z', b'Z'])
        self.serializer.download(response, image)
        self.assertEqual(b'ZZZ', response.body)
        self.assertEqual('3', response.headers['Content-Length'])
        self.assertNotIn('Content-MD5', response.headers)
        self.assertEqual('application/octet-stream',
                         response.headers['Content-Type'])

    def test_range_requests_for_image_downloads(self):
        """
        Test partial download 'Range' requests for images (random image access)
        """
        def download_successful_Range(d_range):
            request = wsgi.Request.blank('/')
            request.environ = {}
            request.headers['Range'] = d_range
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=[b'X', b'Y', b'Z'])
            self.serializer.download(response, image)
            self.assertEqual(206, response.status_code)
            self.assertEqual('2', response.headers['Content-Length'])
            self.assertEqual('bytes 1-2/3', response.headers['Content-Range'])
            self.assertEqual(b'YZ', response.body)

        download_successful_Range('bytes=1-2')
        download_successful_Range('bytes=1-')
        download_successful_Range('bytes=1-3')
        download_successful_Range('bytes=-2')
        download_successful_Range('bytes=1-100')

        def full_image_download_w_range(d_range):
            request = wsgi.Request.blank('/')
            request.environ = {}
            request.headers['Range'] = d_range
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=[b'X', b'Y', b'Z'])
            self.serializer.download(response, image)
            self.assertEqual(206, response.status_code)
            self.assertEqual('3', response.headers['Content-Length'])
            self.assertEqual('bytes 0-2/3', response.headers['Content-Range'])
            self.assertEqual(b'XYZ', response.body)

        full_image_download_w_range('bytes=0-')
        full_image_download_w_range('bytes=0-2')
        full_image_download_w_range('bytes=0-3')
        full_image_download_w_range('bytes=-3')
        full_image_download_w_range('bytes=-4')
        full_image_download_w_range('bytes=0-100')
        full_image_download_w_range('bytes=-100')

        def download_failures_Range(d_range):
            request = wsgi.Request.blank('/')
            request.environ = {}
            request.headers['Range'] = d_range
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=[b'Z', b'Z', b'Z'])
            self.assertRaises(webob.exc.HTTPRequestRangeNotSatisfiable,
                              self.serializer.download,
                              response, image)
            return

        download_failures_Range('bytes=4-1')
        download_failures_Range('bytes=4-')
        download_failures_Range('bytes=3-')
        download_failures_Range('bytes=1')
        download_failures_Range('bytes=100')
        download_failures_Range('bytes=100-')
        download_failures_Range('bytes=')

    def test_multi_range_requests_raises_bad_request_error(self):
        request = wsgi.Request.blank('/')
        request.environ = {}
        request.headers['Range'] = 'bytes=0-0,-1'
        response = webob.Response()
        response.request = request
        image = FakeImage(size=3, data=[b'Z', b'Z', b'Z'])
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.serializer.download,
                          response, image)

    def test_download_failure_with_valid_range(self):
        with mock.patch.object(glance.domain.proxy.Image,
                               'get_data') as mock_get_data:
            mock_get_data.side_effect = glance_store.NotFound(image="image")
        request = wsgi.Request.blank('/')
        request.environ = {}
        request.headers['Range'] = 'bytes=1-2'
        response = webob.Response()
        response.request = request
        image = FakeImage(size=3, data=[b'Z', b'Z', b'Z'])
        image.get_data = mock_get_data
        self.assertRaises(webob.exc.HTTPNoContent,
                          self.serializer.download,
                          response, image)

    def test_content_range_requests_for_image_downloads(self):
        """
        Even though Content-Range is incorrect on requests, we support it
        for backward compatibility with clients written for pre-Pike
        Glance.
        The following test is for 'Content-Range' requests, which we have
        to ensure that we prevent regression.
        """
        def download_successful_ContentRange(d_range):
            request = wsgi.Request.blank('/')
            request.environ = {}
            request.headers['Content-Range'] = d_range
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=[b'X', b'Y', b'Z'])
            self.serializer.download(response, image)
            self.assertEqual(206, response.status_code)
            self.assertEqual('2', response.headers['Content-Length'])
            self.assertEqual('bytes 1-2/3', response.headers['Content-Range'])
            self.assertEqual(b'YZ', response.body)

        download_successful_ContentRange('bytes 1-2/3')
        download_successful_ContentRange('bytes 1-2/*')

        def download_failures_ContentRange(d_range):
            request = wsgi.Request.blank('/')
            request.environ = {}
            request.headers['Content-Range'] = d_range
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=[b'Z', b'Z', b'Z'])
            self.assertRaises(webob.exc.HTTPRequestRangeNotSatisfiable,
                              self.serializer.download,
                              response, image)
            return

        download_failures_ContentRange('bytes -3/3')
        download_failures_ContentRange('bytes 1-/3')
        download_failures_ContentRange('bytes 1-3/3')
        download_failures_ContentRange('bytes 1-4/3')
        download_failures_ContentRange('bytes 1-4/*')
        download_failures_ContentRange('bytes 4-1/3')
        download_failures_ContentRange('bytes 4-1/*')
        download_failures_ContentRange('bytes 4-8/*')
        download_failures_ContentRange('bytes 4-8/10')
        download_failures_ContentRange('bytes 4-8/3')

    def test_download_failure_with_valid_content_range(self):
        with mock.patch.object(glance.domain.proxy.Image,
                               'get_data') as mock_get_data:
            mock_get_data.side_effect = glance_store.NotFound(image="image")
        request = wsgi.Request.blank('/')
        request.environ = {}
        request.headers['Content-Range'] = 'bytes %s-%s/3' % (1, 2)
        response = webob.Response()
        response.request = request
        image = FakeImage(size=3, data=[b'Z', b'Z', b'Z'])
        image.get_data = mock_get_data
        self.assertRaises(webob.exc.HTTPNoContent,
                          self.serializer.download,
                          response, image)

    def test_download_with_checksum(self):
        request = wsgi.Request.blank('/')
        request.environ = {}
        response = webob.Response()
        response.request = request
        checksum = '0745064918b49693cca64d6b6a13d28a'
        image = FakeImage(size=3, checksum=checksum, data=[b'Z', b'Z', b'Z'])
        self.serializer.download(response, image)
        self.assertEqual(b'ZZZ', response.body)
        self.assertEqual('3', response.headers['Content-Length'])
        self.assertEqual(checksum, response.headers['Content-MD5'])
        self.assertEqual('application/octet-stream',
                         response.headers['Content-Type'])

    def test_download_forbidden(self):
        """Make sure the serializer can return 403 forbidden error instead of
        500 internal server error.
        """
        def get_data(*args, **kwargs):
            raise exception.Forbidden()

        self.mock_object(glance.domain.proxy.Image,
                         'get_data',
                         get_data)
        request = wsgi.Request.blank('/')
        request.environ = {}
        response = webob.Response()
        response.request = request
        image = FakeImage(size=3, data=iter('ZZZ'))
        image.get_data = get_data
        self.assertRaises(webob.exc.HTTPForbidden,
                          self.serializer.download,
                          response, image)

    def test_download_no_content(self):
        """Test image download returns HTTPNoContent

        Make sure that serializer returns 204 no content error in case of
        image data is not available at specified location.
        """
        with mock.patch.object(glance.domain.proxy.Image,
                               'get_data') as mock_get_data:
            mock_get_data.side_effect = glance_store.NotFound(image="image")

            request = wsgi.Request.blank('/')
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=iter('ZZZ'))
            image.get_data = mock_get_data
            self.assertRaises(webob.exc.HTTPNoContent,
                              self.serializer.download,
                              response, image)

    def test_download_service_unavailable(self):
        """Test image download returns HTTPServiceUnavailable."""
        with mock.patch.object(glance.domain.proxy.Image,
                               'get_data') as mock_get_data:
            mock_get_data.side_effect = glance_store.RemoteServiceUnavailable()

            request = wsgi.Request.blank('/')
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=iter('ZZZ'))
            image.get_data = mock_get_data
            self.assertRaises(webob.exc.HTTPServiceUnavailable,
                              self.serializer.download,
                              response, image)

    def test_download_store_get_not_support(self):
        """Test image download returns HTTPBadRequest.

        Make sure that serializer returns 400 bad request error in case of
        getting images from this store is not supported at specified location.
        """
        with mock.patch.object(glance.domain.proxy.Image,
                               'get_data') as mock_get_data:
            mock_get_data.side_effect = glance_store.StoreGetNotSupported()

            request = wsgi.Request.blank('/')
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=iter('ZZZ'))
            image.get_data = mock_get_data
            self.assertRaises(webob.exc.HTTPBadRequest,
                              self.serializer.download,
                              response, image)

    def test_download_store_random_get_not_support(self):
        """Test image download returns HTTPBadRequest.

        Make sure that serializer returns 400 bad request error in case of
        getting randomly images from this store is not supported at
        specified location.
        """
        with mock.patch.object(glance.domain.proxy.Image,
                               'get_data') as m_get_data:
            err = glance_store.StoreRandomGetNotSupported(offset=0,
                                                          chunk_size=0)
            m_get_data.side_effect = err

            request = wsgi.Request.blank('/')
            response = webob.Response()
            response.request = request
            image = FakeImage(size=3, data=iter('ZZZ'))
            image.get_data = m_get_data
            self.assertRaises(webob.exc.HTTPBadRequest,
                              self.serializer.download,
                              response, image)

    def test_upload(self):
        request = webob.Request.blank('/')
        request.environ = {}
        response = webob.Response()
        response.request = request
        self.serializer.upload(response, {})
        self.assertEqual(http.NO_CONTENT, response.status_int)
        self.assertEqual('0', response.headers['Content-Length'])

    def test_stage(self):
        request = webob.Request.blank('/')
        request.environ = {}
        response = webob.Response()
        response.request = request
        self.serializer.stage(response, {})
        self.assertEqual(http.NO_CONTENT, response.status_int)
        self.assertEqual('0', response.headers['Content-Length'])


class TestMultiBackendImagesController(base.MultiStoreClearingUnitTest):

    def setUp(self):
        super(TestMultiBackendImagesController, self).setUp()

        self.config(debug=True)
        self.image_repo = FakeImageRepo()
        db = unit_test_utils.FakeDB()
        policy = unit_test_utils.FakePolicyEnforcer()
        notifier = unit_test_utils.FakeNotifier()
        store = unit_test_utils.FakeStoreAPI()
        self.controller = glance.api.v2.image_data.ImageDataController()
        self.controller.gateway = FakeGateway(db, store, notifier, policy,
                                              self.image_repo)
        # FIXME(abhishekk): Everything is fake in this test, so mocked the
        # image muntable_check, Later we need to fix these tests to use
        # some realistic data
        patcher = mock.patch('glance.api.v2.policy.check_is_image_mutable')
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_upload(self):
        request = unit_test_utils.get_fake_request(roles=['admin', 'member'])
        image = FakeImage('abcd')
        self.image_repo.result = image
        self.controller.upload(request, unit_test_utils.UUID2, 'YYYY', 4)
        self.assertEqual('YYYY', image.data)
        self.assertEqual(4, image.size)

    def test_upload_invalid_backend_in_request_header(self):
        request = unit_test_utils.get_fake_request()
        request.headers['x-image-meta-store'] = 'dummy'
        image = FakeImage('abcd')
        self.image_repo.result = image
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.upload,
                          request, unit_test_utils.UUID2, 'YYYY', 4)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
SQLAlchemy models for glance data
"""

import uuid

from oslo_db.sqlalchemy import models
from oslo_serialization import jsonutils
from sqlalchemy import BigInteger
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Enum
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import backref, relationship
from sqlalchemy import sql
from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy.types import TypeDecorator
from sqlalchemy import UniqueConstraint

from glance.common import timeutils


BASE = declarative_base()


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string"""

    impl = Text

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = jsonutils.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = jsonutils.loads(value)
        return value


class GlanceBase(models.ModelBase, models.TimestampMixin):
    """Base class for Glance Models."""

    __table_args__ = {'mysql_engine': 'InnoDB', 'mysql_charset': 'utf8'}
    __table_initialized__ = False
    __protected_attributes__ = set([
        "created_at", "updated_at", "deleted_at", "deleted"])

    def save(self, session=None):
        from glance.db.sqlalchemy import api as db_api
        super(GlanceBase, self).save(session or db_api.get_session())

    created_at = Column(DateTime, default=lambda: timeutils.utcnow(),
                        nullable=False)
    # TODO(vsergeyev): Column `updated_at` have no default value in
    #                  OpenStack common code. We should decide, is this value
    #                  required and make changes in oslo (if required) or
    #                  in glance (if not).
    updated_at = Column(DateTime, default=lambda: timeutils.utcnow(),
                        nullable=True, onupdate=lambda: timeutils.utcnow())
    # TODO(boris-42): Use SoftDeleteMixin instead of deleted Column after
    #                 migration that provides UniqueConstraints and change
    #                 type of this column.
    deleted_at = Column(DateTime)
    deleted = Column(Boolean, nullable=False, default=False)

    def delete(self, session=None):
        """Delete this object."""
        self.deleted = True
        self.deleted_at = timeutils.utcnow()
        self.save(session=session)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def to_dict(self):
        d = self.__dict__.copy()
        # NOTE(flaper87): Remove
        # private state instance
        # It is not serializable
        # and causes CircularReference
        d.pop("_sa_instance_state")
        return d


class Image(BASE, GlanceBase):
    """Represents an image in the datastore."""
    __tablename__ = 'images'
    __table_args__ = (Index('checksum_image_idx', 'checksum'),
                      Index('visibility_image_idx', 'visibility'),
                      Index('ix_images_deleted', 'deleted'),
                      Index('owner_image_idx', 'owner'),
                      Index('created_at_image_idx', 'created_at'),
                      Index('updated_at_image_idx', 'updated_at'),
                      Index('os_hidden_image_idx', 'os_hidden'),
                      Index('os_hash_value_image_idx', 'os_hash_value'))

    id = Column(String(36), primary_key=True,
                default=lambda: str(uuid.uuid4()))
    name = Column(String(255))
    disk_format = Column(String(20))
    container_format = Column(String(20))
    size = Column(BigInteger().with_variant(Integer, "sqlite"))
    virtual_size = Column(BigInteger().with_variant(Integer, "sqlite"))
    status = Column(String(30), nullable=False)
    visibility = Column(Enum('private', 'public', 'shared', 'community',
                        name='image_visibility'), nullable=False,
                        server_default='shared')
    checksum = Column(String(32))
    os_hash_algo = Column(String(64))
    os_hash_value = Column(String(128))
    min_disk = Column(Integer, nullable=False, default=0)
    min_ram = Column(Integer, nullable=False, default=0)
    owner = Column(String(255))
    protected = Column(Boolean, nullable=False, default=False,
                       server_default=sql.expression.false())
    os_hidden = Column(Boolean, nullable=False, default=False,
                       server_default=sql.expression.false())


class ImageProperty(BASE, GlanceBase):
    """Represents an image properties in the datastore."""
    __tablename__ = 'image_properties'
    __table_args__ = (Index('ix_image_properties_image_id', 'image_id'),
                      Index('ix_image_properties_deleted', 'deleted'),
                      UniqueConstraint('image_id',
                                       'name',
                                       name='ix_image_properties_'
                                            'image_id_name'),)

    id = Column(Integer, primary_key=True)
    image_id = Column(String(36), ForeignKey('images.id'),
                      nullable=False)
    image = relationship(Image, backref=backref('properties'))

    name = Column(String(255), nullable=False)
    value = Column(Text)


class ImageTag(BASE, GlanceBase):
    """Represents an image tag in the datastore."""
    __tablename__ = 'image_tags'
    __table_args__ = (Index('ix_image_tags_image_id', 'image_id'),
                      Index('ix_image_tags_image_id_tag_value',
                            'image_id',
                            'value'),)

    id = Column(Integer, primary_key=True, nullable=False)
    image_id = Column(String(36), ForeignKey('images.id'), nullable=False)
    image = relationship(Image, backref=backref('tags'))
    value = Column(String(255), nullable=False)


class ImageLocation(BASE, GlanceBase):
    """Represents an image location in the datastore."""
    __tablename__ = 'image_locations'
    __table_args__ = (Index('ix_image_locations_image_id', 'image_id'),
                      Index('ix_image_locations_deleted', 'deleted'),)

    id = Column(Integer, primary_key=True, nullable=False)
    image_id = Column(String(36), ForeignKey('images.id'), nullable=False)
    image = relationship(Image, backref=backref('locations'))
    value = Column(Text(), nullable=False)
    meta_data = Column(JSONEncodedDict(), default={})
    status = Column(String(30), server_default='active', nullable=False)


class ImageMember(BASE, GlanceBase):
    """Represents an image members in the datastore."""
    __tablename__ = 'image_members'
    unique_constraint_key_name = 'image_members_image_id_member_deleted_at_key'
    __table_args__ = (Index('ix_image_members_deleted', 'deleted'),
                      Index('ix_image_members_image_id', 'image_id'),
                      Index('ix_image_members_image_id_member',
                            'image_id',
                            'member'),
                      UniqueConstraint('image_id',
                                       'member',
                                       'deleted_at',
                                       name=unique_constraint_key_name),)

    id = Column(Integer, primary_key=True)
    image_id = Column(String(36), ForeignKey('images.id'),
                      nullable=False)
    image = relationship(Image, backref=backref('members'))

    member = Column(String(255), nullable=False)
    can_share = Column(Boolean, nullable=False, default=False)
    status = Column(String(20), nullable=False, default="pending",
                    server_default='pending')


class Task(BASE, GlanceBase):
    """Represents an task in the datastore"""
    __tablename__ = 'tasks'
    __table_args__ = (Index('ix_tasks_type', 'type'),
                      Index('ix_tasks_status', 'status'),
                      Index('ix_tasks_owner', 'owner'),
                      Index('ix_tasks_deleted', 'deleted'),
                      Index('ix_tasks_image_id', 'image_id'),
                      Index('ix_tasks_updated_at', 'updated_at'))

    id = Column(String(36), primary_key=True,
                default=lambda: str(uuid.uuid4()))
    type = Column(String(30), nullable=False)
    status = Column(String(30), nullable=False)
    owner = Column(String(255), nullable=False)
    expires_at = Column(DateTime, nullable=True)
    image_id = Column(String(36), nullable=True)
    request_id = Column(String(64), nullable=True)
    user_id = Column(String(64), nullable=True)


class TaskInfo(BASE, models.ModelBase):
    """Represents task info in the datastore"""
    __tablename__ = 'task_info'

    task_id = Column(String(36),
                     ForeignKey('tasks.id'),
                     primary_key=True,
                     nullable=False)

    task = relationship(Task, backref=backref('info', uselist=False))

    # NOTE(nikhil): input and result are stored as text in the DB.
    # SQLAlchemy marshals the data to/from JSON using custom type
    # JSONEncodedDict. It uses simplejson underneath.
    input = Column(JSONEncodedDict())
    result = Column(JSONEncodedDict())
    message = Column(Text)


class NodeReference(BASE, models.ModelBase):
    """Represents node info in the datastore"""
    __tablename__ = 'node_reference'
    __table_args__ = (UniqueConstraint(
        'node_reference_url',
        name='uq_node_reference_node_reference_url'),)

    node_reference_id = Column(BigInteger().with_variant(Integer, 'sqlite'),
                               primary_key=True,
                               nullable=False, autoincrement=True)
    node_reference_url = Column(String(length=255),
                                nullable=False)


class CachedImages(BASE, models.ModelBase):
    """Represents an image tag in the datastore."""
    __tablename__ = 'cached_images'
    __table_args__ = (UniqueConstraint(
        'image_id',
        'node_reference_id',
        name='ix_cached_images_image_id_node_reference_id'),)

    id = Column(BigInteger().with_variant(Integer, 'sqlite'),
                primary_key=True, autoincrement=True,
                nullable=False)
    image_id = Column(String(36), nullable=False)
    last_accessed = Column(DateTime, nullable=False)
    last_modified = Column(DateTime, nullable=False)
    size = Column(BigInteger(), nullable=False)
    hits = Column(Integer, nullable=False)
    checksum = Column(String(32), nullable=True)
    node_reference_id = Column(
        BigInteger().with_variant(Integer, 'sqlite'),
        ForeignKey('node_reference.node_reference_id'),
        nullable=False)


def register_models(engine):
    """Create database tables for all models with the given engine."""
    models = (Image, ImageProperty, ImageMember)
    for model in models:
        model.metadata.create_all(engine)


def unregister_models(engine):
    """Drop database tables for all models with the given engine."""
    models = (Image, ImageProperty)
    for model in models:
        model.metadata.drop_all(engine)



# Copyright (C) 2023 RedHat Inc
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""adds cache_node_reference and cached_images table(s)

Revision ID: 2024_1_expand01
Revises: 2023_1_expand01
Create Date: 2023-10-31 11:55:16.657499

"""

from alembic import op
from sqlalchemy.schema import (
    Column, PrimaryKeyConstraint, ForeignKeyConstraint, UniqueConstraint)

from glance.db.sqlalchemy.schema import (
    Integer, BigInteger, DateTime, String)  # noqa

# revision identifiers, used by Alembic.
revision = '2024_1_expand01'
down_revision = '2023_1_expand01'
branch_labels = None
depends_on = None


def _add_node_reference_table():
    op.create_table('node_reference',
                    Column('node_reference_id',
                           BigInteger().with_variant(Integer, 'sqlite'),
                           nullable=False,
                           autoincrement=True),
                    Column('node_reference_url', String(length=255),
                           nullable=False),
                    PrimaryKeyConstraint('node_reference_id'),
                    UniqueConstraint(
                        'node_reference_url',
                        name='uq_node_reference_node_reference_url'),
                    mysql_engine='InnoDB',
                    mysql_charset='utf8',
                    extend_existing=True)


def _add_cached_images_table():
    op.create_table('cached_images',
                    Column('id', BigInteger().with_variant(Integer, 'sqlite'),
                           autoincrement=True,
                           nullable=False),
                    Column('image_id', String(length=36), nullable=False),
                    Column('last_accessed', DateTime(), nullable=False),
                    Column('last_modified', DateTime(), nullable=False),
                    Column('size', BigInteger(), nullable=False),
                    Column('hits', Integer(), nullable=False),
                    Column('checksum', String(length=32), nullable=True),
                    Column('node_reference_id',
                           BigInteger().with_variant(Integer, 'sqlite'),
                           nullable=False),
                    PrimaryKeyConstraint('id'),
                    ForeignKeyConstraint(
                        ['node_reference_id'],
                        ['node_reference.node_reference_id'], ),
                    UniqueConstraint(
                        'image_id',
                        'node_reference_id',
                        name='ix_cached_images_image_id_node_reference_id'),
                    mysql_engine='InnoDB',
                    mysql_charset='utf8',
                    extend_existing=True)


def upgrade():
    _add_node_reference_table()
    _add_cached_images_table()



# Copyright 2012 OpenStack Foundation.
# Copyright 2014 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import encodeutils

from glance.common import exception
import glance.context
import glance.db
import glance.tests.unit.utils as unit_test_utils
import glance.tests.utils as test_utils

TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'
TENANT2 = '2c014f32-55eb-467d-8fcb-4bd706012f81'
TENANT3 = '5a3e60e8-cfa9-4a9e-a90a-62b42cea92b8'
TENANT4 = 'c6c87f25-8a94-47ed-8c83-053c25f42df4'

USER1 = '54492ba0-f4df-4e4e-be62-27f4d76b29cf'

NAMESPACE1 = 'namespace1'
NAMESPACE2 = 'namespace2'
NAMESPACE3 = 'namespace3'
NAMESPACE4 = 'namespace4'

PROPERTY1 = 'Property1'
PROPERTY2 = 'Property2'
PROPERTY3 = 'Property3'

OBJECT1 = 'Object1'
OBJECT2 = 'Object2'
OBJECT3 = 'Object3'

TAG1 = 'Tag1'
TAG2 = 'Tag2'
TAG3 = 'Tag3'
TAG4 = 'Tag4'
TAG5 = 'Tag5'

RESOURCE_TYPE1 = 'ResourceType1'
RESOURCE_TYPE2 = 'ResourceType2'
RESOURCE_TYPE3 = 'ResourceType3'


def _db_namespace_fixture(**kwargs):
    namespace = {
        'namespace': None,
        'display_name': None,
        'description': None,
        'visibility': True,
        'protected': False,
        'owner': None
    }
    namespace.update(kwargs)
    return namespace


def _db_property_fixture(name, **kwargs):
    property = {
        'name': name,
        'json_schema': {"type": "string", "title": "title"},
    }
    property.update(kwargs)
    return property


def _db_object_fixture(name, **kwargs):
    obj = {
        'name': name,
        'description': None,
        'json_schema': {},
        'required': '[]',
    }
    obj.update(kwargs)
    return obj


def _db_tag_fixture(name, **kwargs):
    obj = {
        'name': name
    }
    obj.update(kwargs)
    return obj


def _db_tags_fixture(names=None):
    tags = []
    if names:
        tag_name_list = names
    else:
        tag_name_list = [TAG1, TAG2, TAG3]

    for tag_name in tag_name_list:
        tags.append(_db_tag_fixture(tag_name))
    return tags


def _db_resource_type_fixture(name, **kwargs):
    obj = {
        'name': name,
        'protected': False,
    }
    obj.update(kwargs)
    return obj


def _db_namespace_resource_type_fixture(name, **kwargs):
    obj = {
        'name': name,
        'properties_target': None,
        'prefix': None,
    }
    obj.update(kwargs)
    return obj


class TestMetadefRepo(test_utils.BaseTestCase):

    def setUp(self):
        super(TestMetadefRepo, self).setUp()
        self.db = unit_test_utils.FakeDB(initialize=False)
        self.context = glance.context.RequestContext(user=USER1,
                                                     tenant=TENANT1)
        self.namespace_repo = glance.db.MetadefNamespaceRepo(self.context,
                                                             self.db)
        self.property_repo = glance.db.MetadefPropertyRepo(self.context,
                                                           self.db)
        self.object_repo = glance.db.MetadefObjectRepo(self.context,
                                                       self.db)
        self.tag_repo = glance.db.MetadefTagRepo(self.context,
                                                 self.db)
        self.resource_type_repo = glance.db.MetadefResourceTypeRepo(
            self.context, self.db)
        self.namespace_factory = glance.domain.MetadefNamespaceFactory()
        self.property_factory = glance.domain.MetadefPropertyFactory()
        self.object_factory = glance.domain.MetadefObjectFactory()
        self.tag_factory = glance.domain.MetadefTagFactory()
        self.resource_type_factory = glance.domain.MetadefResourceTypeFactory()
        self._create_namespaces()
        self._create_properties()
        self._create_objects()
        self._create_tags()
        self._create_resource_types()

    def _create_namespaces(self):
        self.namespaces = [
            _db_namespace_fixture(namespace=NAMESPACE1,
                                  display_name='1',
                                  description='desc1',
                                  visibility='private',
                                  protected=True,
                                  owner=TENANT1),
            _db_namespace_fixture(namespace=NAMESPACE2,
                                  display_name='2',
                                  description='desc2',
                                  visibility='public',
                                  protected=False,
                                  owner=TENANT1),
            _db_namespace_fixture(namespace=NAMESPACE3,
                                  display_name='3',
                                  description='desc3',
                                  visibility='private',
                                  protected=True,
                                  owner=TENANT3),
            _db_namespace_fixture(namespace=NAMESPACE4,
                                  display_name='4',
                                  description='desc4',
                                  visibility='public',
                                  protected=True,
                                  owner=TENANT3)
        ]
        [self.db.metadef_namespace_create(None, namespace)
         for namespace in self.namespaces]

    def _create_properties(self):
        self.properties = [
            _db_property_fixture(name=PROPERTY1),
            _db_property_fixture(name=PROPERTY2),
            _db_property_fixture(name=PROPERTY3)
        ]
        [self.db.metadef_property_create(self.context, NAMESPACE1, property)
         for property in self.properties]
        [self.db.metadef_property_create(self.context, NAMESPACE4, property)
         for property in self.properties]

    def _create_objects(self):
        self.objects = [
            _db_object_fixture(name=OBJECT1,
                               description='desc1'),
            _db_object_fixture(name=OBJECT2,
                               description='desc2'),
            _db_object_fixture(name=OBJECT3,
                               description='desc3'),
        ]
        [self.db.metadef_object_create(self.context, NAMESPACE1, object)
         for object in self.objects]
        [self.db.metadef_object_create(self.context, NAMESPACE4, object)
         for object in self.objects]

    def _create_tags(self):
        self.tags = [
            _db_tag_fixture(name=TAG1),
            _db_tag_fixture(name=TAG2),
            _db_tag_fixture(name=TAG3),
        ]
        [self.db.metadef_tag_create(self.context, NAMESPACE1, tag)
         for tag in self.tags]
        [self.db.metadef_tag_create(self.context, NAMESPACE4, tag)
         for tag in self.tags]

    def _create_resource_types(self):
        self.resource_types = [
            _db_resource_type_fixture(name=RESOURCE_TYPE1,
                                      protected=False),
            _db_resource_type_fixture(name=RESOURCE_TYPE2,
                                      protected=False),
            _db_resource_type_fixture(name=RESOURCE_TYPE3,
                                      protected=True),
        ]
        [self.db.metadef_resource_type_create(self.context, resource_type)
         for resource_type in self.resource_types]

    def test_get_namespace(self):
        namespace = self.namespace_repo.get(NAMESPACE1)
        self.assertEqual(NAMESPACE1, namespace.namespace)
        self.assertEqual('desc1', namespace.description)
        self.assertEqual('1', namespace.display_name)
        self.assertEqual(TENANT1, namespace.owner)
        self.assertTrue(namespace.protected)
        self.assertEqual('private', namespace.visibility)

    def test_get_namespace_not_found(self):
        fake_namespace = "fake_namespace"
        exc = self.assertRaises(exception.NotFound,
                                self.namespace_repo.get,
                                fake_namespace)
        self.assertIn(fake_namespace, encodeutils.exception_to_unicode(exc))

    def test_get_namespace_forbidden(self):
        self.assertRaises(exception.MetadefForbidden,
                          self.namespace_repo.get,
                          NAMESPACE3)

    def test_list_namespace(self):
        namespaces = self.namespace_repo.list()
        namespace_names = set([n.namespace for n in namespaces])
        self.assertEqual(set([NAMESPACE1, NAMESPACE2, NAMESPACE4]),
                         namespace_names)

    def test_list_private_namespaces(self):
        filters = {'visibility': 'private'}
        namespaces = self.namespace_repo.list(filters=filters)
        namespace_names = set([n.namespace for n in namespaces])
        self.assertEqual(set([NAMESPACE1]), namespace_names)

    def test_add_namespace(self):
        # NOTE(pawel-koniszewski): Change db_namespace_fixture to
        # namespace_factory when namespace primary key in DB
        # will be changed from Integer to UUID
        namespace = _db_namespace_fixture(namespace='added_namespace',
                                          display_name='fake',
                                          description='fake_desc',
                                          visibility='public',
                                          protected=True,
                                          owner=TENANT1)
        self.assertEqual('added_namespace', namespace['namespace'])
        self.db.metadef_namespace_create(None, namespace)
        retrieved_namespace = self.namespace_repo.get(namespace['namespace'])
        self.assertEqual('added_namespace', retrieved_namespace.namespace)

    def test_save_namespace(self):
        namespace = self.namespace_repo.get(NAMESPACE1)
        namespace.display_name = 'save_name'
        namespace.description = 'save_desc'
        self.namespace_repo.save(namespace)
        namespace = self.namespace_repo.get(NAMESPACE1)
        self.assertEqual('save_name', namespace.display_name)
        self.assertEqual('save_desc', namespace.description)

    def test_remove_namespace(self):
        namespace = self.namespace_repo.get(NAMESPACE1)
        self.namespace_repo.remove(namespace)
        self.assertRaises(exception.NotFound, self.namespace_repo.get,
                          NAMESPACE1)

    def test_remove_namespace_not_found(self):
        fake_name = 'fake_name'
        namespace = self.namespace_repo.get(NAMESPACE1)
        namespace.namespace = fake_name
        exc = self.assertRaises(exception.NotFound, self.namespace_repo.remove,
                                namespace)
        self.assertIn(fake_name, encodeutils.exception_to_unicode(exc))

    def test_get_property(self):
        property = self.property_repo.get(NAMESPACE1, PROPERTY1)
        namespace = self.namespace_repo.get(NAMESPACE1)
        self.assertEqual(PROPERTY1, property.name)
        self.assertEqual(namespace.namespace, property.namespace.namespace)

    def test_get_property_not_found(self):
        exc = self.assertRaises(exception.NotFound,
                                self.property_repo.get,
                                NAMESPACE2, PROPERTY1)
        self.assertIn(PROPERTY1, encodeutils.exception_to_unicode(exc))

    def test_list_property(self):
        properties = self.property_repo.list(filters={'namespace': NAMESPACE1})
        property_names = set([p.name for p in properties])
        self.assertEqual(set([PROPERTY1, PROPERTY2, PROPERTY3]),
                         property_names)

    def test_list_property_empty_result(self):
        properties = self.property_repo.list(filters={'namespace': NAMESPACE2})
        property_names = set([p.name for p in properties])
        self.assertEqual(set([]),
                         property_names)

    def test_list_property_namespace_not_found(self):
        exc = self.assertRaises(exception.NotFound, self.property_repo.list,
                                filters={'namespace': 'not-a-namespace'})
        self.assertIn('not-a-namespace', encodeutils.exception_to_unicode(exc))

    def test_add_property(self):
        # NOTE(pawel-koniszewski): Change db_property_fixture to
        # property_factory when property primary key in DB
        # will be changed from Integer to UUID
        property = _db_property_fixture(name='added_property')
        self.assertEqual('added_property', property['name'])
        self.db.metadef_property_create(self.context, NAMESPACE1, property)
        retrieved_property = self.property_repo.get(NAMESPACE1,
                                                    'added_property')
        self.assertEqual('added_property', retrieved_property.name)

    def test_add_property_namespace_forbidden(self):
        # NOTE(pawel-koniszewski): Change db_property_fixture to
        # property_factory when property primary key in DB
        # will be changed from Integer to UUID
        property = _db_property_fixture(name='added_property')
        self.assertEqual('added_property', property['name'])
        self.assertRaises(exception.Forbidden, self.db.metadef_property_create,
                          self.context, NAMESPACE3, property)

    def test_add_property_namespace_not_found(self):
        # NOTE(pawel-koniszewski): Change db_property_fixture to
        # property_factory when property primary key in DB
        # will be changed from Integer to UUID
        property = _db_property_fixture(name='added_property')
        self.assertEqual('added_property', property['name'])
        self.assertRaises(exception.NotFound, self.db.metadef_property_create,
                          self.context, 'not_a_namespace', property)

    def test_save_property(self):
        property = self.property_repo.get(NAMESPACE1, PROPERTY1)
        property.schema = '{"save": "schema"}'
        self.property_repo.save(property)
        property = self.property_repo.get(NAMESPACE1, PROPERTY1)
        self.assertEqual(PROPERTY1, property.name)
        self.assertEqual('{"save": "schema"}', property.schema)

    def test_remove_property(self):
        property = self.property_repo.get(NAMESPACE1, PROPERTY1)
        self.property_repo.remove(property)
        self.assertRaises(exception.NotFound, self.property_repo.get,
                          NAMESPACE1, PROPERTY1)

    def test_remove_property_not_found(self):
        fake_name = 'fake_name'
        property = self.property_repo.get(NAMESPACE1, PROPERTY1)
        property.name = fake_name
        self.assertRaises(exception.NotFound, self.property_repo.remove,
                          property)

    def test_get_object(self):
        object = self.object_repo.get(NAMESPACE1, OBJECT1)
        namespace = self.namespace_repo.get(NAMESPACE1)
        self.assertEqual(OBJECT1, object.name)
        self.assertEqual('desc1', object.description)
        self.assertEqual(['[]'], object.required)
        self.assertEqual({}, object.properties)
        self.assertEqual(namespace.namespace, object.namespace.namespace)

    def test_get_object_not_found(self):
        exc = self.assertRaises(exception.NotFound, self.object_repo.get,
                                NAMESPACE2, OBJECT1)
        self.assertIn(OBJECT1, encodeutils.exception_to_unicode(exc))

    def test_list_object(self):
        objects = self.object_repo.list(filters={'namespace': NAMESPACE1})
        object_names = set([o.name for o in objects])
        self.assertEqual(set([OBJECT1, OBJECT2, OBJECT3]), object_names)

    def test_list_object_empty_result(self):
        objects = self.object_repo.list(filters={'namespace': NAMESPACE2})
        object_names = set([o.name for o in objects])
        self.assertEqual(set([]), object_names)

    def test_list_object_namespace_not_found(self):
        exc = self.assertRaises(exception.NotFound, self.object_repo.list,
                                filters={'namespace': 'not-a-namespace'})
        self.assertIn('not-a-namespace', encodeutils.exception_to_unicode(exc))

    def test_add_object(self):
        # NOTE(pawel-koniszewski): Change db_object_fixture to
        # object_factory when object primary key in DB
        # will be changed from Integer to UUID
        object = _db_object_fixture(name='added_object')
        self.assertEqual('added_object', object['name'])
        self.db.metadef_object_create(self.context, NAMESPACE1, object)
        retrieved_object = self.object_repo.get(NAMESPACE1,
                                                'added_object')
        self.assertEqual('added_object', retrieved_object.name)

    def test_add_object_namespace_forbidden(self):
        # NOTE(pawel-koniszewski): Change db_object_fixture to
        # object_factory when object primary key in DB
        # will be changed from Integer to UUID
        object = _db_object_fixture(name='added_object')
        self.assertEqual('added_object', object['name'])
        self.assertRaises(exception.Forbidden, self.db.metadef_object_create,
                          self.context, NAMESPACE3, object)

    def test_add_object_namespace_not_found(self):
        # NOTE(pawel-koniszewski): Change db_object_fixture to
        # object_factory when object primary key in DB
        # will be changed from Integer to UUID
        object = _db_object_fixture(name='added_object')
        self.assertEqual('added_object', object['name'])
        self.assertRaises(exception.NotFound, self.db.metadef_object_create,
                          self.context, 'not-a-namespace', object)

    def test_save_object(self):
        object = self.object_repo.get(NAMESPACE1, OBJECT1)
        object.required = ['save_req']
        object.description = 'save_desc'
        self.object_repo.save(object)
        object = self.object_repo.get(NAMESPACE1, OBJECT1)
        self.assertEqual(OBJECT1, object.name)
        self.assertEqual(['save_req'], object.required)
        self.assertEqual('save_desc', object.description)

    def test_remove_object(self):
        object = self.object_repo.get(NAMESPACE1, OBJECT1)
        self.object_repo.remove(object)
        self.assertRaises(exception.NotFound, self.object_repo.get,
                          NAMESPACE1, OBJECT1)

    def test_remove_object_not_found(self):
        fake_name = 'fake_name'
        object = self.object_repo.get(NAMESPACE1, OBJECT1)
        object.name = fake_name
        self.assertRaises(exception.NotFound, self.object_repo.remove,
                          object)

    def test_list_resource_type(self):
        resource_type = self.resource_type_repo.list(
            filters={'namespace': NAMESPACE1})
        self.assertEqual(0, len(resource_type))

    def test_get_tag(self):
        tag = self.tag_repo.get(NAMESPACE1, TAG1)
        namespace = self.namespace_repo.get(NAMESPACE1)
        self.assertEqual(TAG1, tag.name)
        self.assertEqual(namespace.namespace, tag.namespace.namespace)

    def test_get_tag_not_found(self):
        exc = self.assertRaises(exception.NotFound, self.tag_repo.get,
                                NAMESPACE2, TAG1)
        self.assertIn(TAG1, encodeutils.exception_to_unicode(exc))

    def test_list_tag(self):
        tags = self.tag_repo.list(filters={'namespace': NAMESPACE1})
        tag_names = set([t.name for t in tags])
        self.assertEqual(set([TAG1, TAG2, TAG3]), tag_names)

    def test_list_tag_empty_result(self):
        tags = self.tag_repo.list(filters={'namespace': NAMESPACE2})
        tag_names = set([t.name for t in tags])
        self.assertEqual(set([]), tag_names)

    def test_list_tag_namespace_not_found(self):
        exc = self.assertRaises(exception.NotFound, self.tag_repo.list,
                                filters={'namespace': 'not-a-namespace'})
        self.assertIn('not-a-namespace', encodeutils.exception_to_unicode(exc))

    def test_add_tag(self):
        # NOTE(pawel-koniszewski): Change db_tag_fixture to
        # tag_factory when tag primary key in DB
        # will be changed from Integer to UUID
        tag = _db_tag_fixture(name='added_tag')
        self.assertEqual('added_tag', tag['name'])
        self.db.metadef_tag_create(self.context, NAMESPACE1, tag)
        retrieved_tag = self.tag_repo.get(NAMESPACE1, 'added_tag')
        self.assertEqual('added_tag', retrieved_tag.name)

    def test_add_tags(self):
        tags = self.tag_repo.list(filters={'namespace': NAMESPACE1})
        tag_names = set([t.name for t in tags])
        self.assertEqual(set([TAG1, TAG2, TAG3]), tag_names)

        tags = _db_tags_fixture([TAG3, TAG4, TAG5])
        self.db.metadef_tag_create_tags(self.context, NAMESPACE1, tags)

        tags = self.tag_repo.list(filters={'namespace': NAMESPACE1})
        tag_names = set([t.name for t in tags])
        self.assertEqual(set([TAG3, TAG4, TAG5]), tag_names)

    def test_add_tags_with_append_true(self):
        tags = self.tag_repo.list(filters={'namespace': NAMESPACE1})
        tag_names = set([t.name for t in tags])
        self.assertEqual(set([TAG1, TAG2, TAG3]), tag_names)

        tags = _db_tags_fixture([TAG4, TAG5])
        self.db.metadef_tag_create_tags(self.context, NAMESPACE1, tags,
                                        can_append=True)
        tags = self.tag_repo.list(filters={'namespace': NAMESPACE1})
        tag_names = set([t.name for t in tags])
        self.assertEqual(set([TAG1, TAG2, TAG3, TAG4, TAG5]), tag_names)

    def test_add_duplicate_tags_with_pre_existing_tags(self):
        tags = self.tag_repo.list(filters={'namespace': NAMESPACE1})
        tag_names = set([t.name for t in tags])
        self.assertEqual(set([TAG1, TAG2, TAG3]), tag_names)

        tags = _db_tags_fixture([TAG5, TAG4, TAG5])
        self.assertRaises(exception.Duplicate,
                          self.db.metadef_tag_create_tags,
                          self.context, NAMESPACE1, tags)

        tags = self.tag_repo.list(filters={'namespace': NAMESPACE1})
        tag_names = set([t.name for t in tags])
        self.assertEqual(set([TAG1, TAG2, TAG3]), tag_names)

    def test_add_tag_namespace_forbidden(self):
        # NOTE(pawel-koniszewski): Change db_tag_fixture to
        # tag_factory when tag primary key in DB
        # will be changed from Integer to UUID
        tag = _db_tag_fixture(name='added_tag')
        self.assertEqual('added_tag', tag['name'])
        self.assertRaises(exception.Forbidden, self.db.metadef_tag_create,
                          self.context, NAMESPACE3, tag)

    def test_add_tag_namespace_not_found(self):
        # NOTE(pawel-koniszewski): Change db_tag_fixture to
        # tag_factory when tag primary key in DB
        # will be changed from Integer to UUID
        tag = _db_tag_fixture(name='added_tag')
        self.assertEqual('added_tag', tag['name'])
        self.assertRaises(exception.NotFound, self.db.metadef_tag_create,
                          self.context, 'not-a-namespace', tag)

    def test_save_tag(self):
        tag = self.tag_repo.get(NAMESPACE1, TAG1)
        self.tag_repo.save(tag)
        tag = self.tag_repo.get(NAMESPACE1, TAG1)
        self.assertEqual(TAG1, tag.name)

    def test_remove_tag(self):
        tag = self.tag_repo.get(NAMESPACE1, TAG1)
        self.tag_repo.remove(tag)
        self.assertRaises(exception.NotFound, self.tag_repo.get,
                          NAMESPACE1, TAG1)

    def test_remove_tag_not_found(self):
        fake_name = 'fake_name'
        tag = self.tag_repo.get(NAMESPACE1, TAG1)
        tag.name = fake_name
        self.assertRaises(exception.NotFound, self.tag_repo.remove, tag)



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import oslo_policy.policy

from glance.api import policy
from glance.tests import functional


GLOBAL_NAMESPACE_DATA = {
    "namespace": "MyNamespace",
    "display_name": "My User Friendly Namespace",
    "description": "My description",
    "resource_type_associations": [{
        "name": "MyResourceType",
        "prefix": "prefix_",
        "properties_target": "temp"
    }],
    "objects": [{
        "name": "MyObject",
        "description": "My object for My namespace",
        "properties": {
            "test_property": {
                "title": "test_property",
                "description": "Test property for My object",
                "type": "string"
            },
        }
    }],
    "tags": [{
        "name": "MyTag",
    }],
    "properties": {
        "TestProperty": {
            "title": "MyTestProperty",
            "description": "Test Property for My namespace",
            "type": "string"
        },
    },
}

NAME_SPACE1 = {
    "namespace": "MyNamespace",
    "display_name": "My User Friendly Namespace",
    "description": "My description"
}

NAME_SPACE2 = {
    "namespace": "MySecondNamespace",
    "display_name": "My User Friendly Namespace",
    "description": "My description"
}


class TestMetadefNamespacesPolicy(functional.SynchronousAPIBase):
    def setUp(self):
        super(TestMetadefNamespacesPolicy, self).setUp()
        self.policy = policy.Enforcer(suppress_deprecation_warnings=True)

    def set_policy_rules(self, rules):
        self.policy.set_rules(
            oslo_policy.policy.Rules.from_dict(rules),
            overwrite=True)

    def start_server(self):
        with mock.patch.object(policy, 'Enforcer') as mock_enf:
            mock_enf.return_value = self.policy
            super(TestMetadefNamespacesPolicy, self).start_server()

    def _verify_forbidden_converted_to_not_found(self, path, method,
                                                 json=None):
        # Note for other reviewers, these tests runs by default using
        # admin role, to test this scenario we need private namespace
        # of current project to be accessed by other projects non-admin
        # user.
        headers = self._headers({
            'X-Tenant-Id': 'fake-tenant-id',
            'X-Roles': 'member',
        })
        resp = self.api_request(method, path, headers=headers, json=json)
        self.assertEqual(404, resp.status_code)

    def test_namespace_list_basic(self):
        self.start_server()
        # First make sure create private namespace works with default policy
        path = '/v2/metadefs/namespaces'
        md_resource = self._create_metadef_resource(path=path,
                                                    data=NAME_SPACE1)
        self.assertEqual('MyNamespace', md_resource['namespace'])

        # First make sure create public namespace works with default policy
        path = '/v2/metadefs/namespaces'
        NAME_SPACE2["visibility"] = 'public'
        md_resource = self._create_metadef_resource(path=path,
                                                    data=NAME_SPACE2)
        self.assertEqual('MySecondNamespace', md_resource['namespace'])

        # Now make sure 'get_metadef_namespaces' allows user to get all the
        # namespaces
        resp = self.api_get(path)
        md_resource = resp.json
        self.assertEqual(2, len(md_resource['namespaces']))

        # Now disable get_metadef_namespaces permissions and make sure any
        # other attempts fail
        self.set_policy_rules({
            'get_metadef_namespaces': '!',
            'get_metadef_namespace': '@',
        })
        resp = self.api_get(path)
        self.assertEqual(403, resp.status_code)

    def test_namespace_list_with_resource_types(self):
        self.start_server()
        # First make sure create namespace works with default policy
        path = '/v2/metadefs/namespaces'
        md_resource = self._create_metadef_resource(path=path,
                                                    data=GLOBAL_NAMESPACE_DATA)
        self.assertEqual('MyNamespace', md_resource['namespace'])

        # Now make sure 'get_metadef_namespaces' allows user to get all the
        # namespaces with associated resource types
        resp = self.api_get(path)
        md_resource = resp.json
        self.assertEqual(1, len(md_resource['namespaces']))
        # Verify that response includes associated resource types as well
        for namespace_obj in md_resource['namespaces']:
            self.assertIn('resource_type_associations', namespace_obj)

        # Now disable list_metadef_resource_types permissions and make sure
        # you get forbidden response
        self.set_policy_rules({
            'get_metadef_namespaces': '@',
            'get_metadef_namespace': '@',
            'list_metadef_resource_types': '!'
        })
        resp = self.api_get(path)
        self.assertEqual(403, resp.status_code)

        # Now enable list_metadef_resource_types and  get_metadef_namespaces
        # permissions and disable get_metadef_namespace permission to make sure
        # you will get empty list as a response
        self.set_policy_rules({
            'get_metadef_namespaces': '@',
            'get_metadef_namespace': '!',
            'list_metadef_resource_types': '@'
        })
        resp = self.api_get(path)
        md_resource = resp.json
        self.assertEqual(0, len(md_resource['namespaces']))
        # Verify that response does not includes associated resource types
        for namespace_obj in md_resource['namespaces']:
            self.assertNotIn('resource_type_associations', namespace_obj)

    def test_namespace_create_basic(self):
        self.start_server()
        # First make sure create namespace works with default policy
        path = '/v2/metadefs/namespaces'
        md_resource = self._create_metadef_resource(path=path,
                                                    data=NAME_SPACE1)
        self.assertEqual('MyNamespace', md_resource['namespace'])

        # Now disable add_metadef_namespace permissions and make sure any other
        # attempts fail
        self.set_policy_rules({
            'add_metadef_namespace': '!',
            'get_metadef_namespace': '@'
        })
        resp = self.api_post(path, json=NAME_SPACE2)
        self.assertEqual(403, resp.status_code)

    def test_namespace_create_with_resource_type_associations(self):
        self.start_server()
        # First make sure you can create namespace and resource type
        # associations with default policy
        path = '/v2/metadefs/namespaces'
        data = {
            "resource_type_associations": [{
                "name": "MyResourceType",
                "prefix": "prefix_",
                "properties_target": "temp"
            }],
        }
        data.update(NAME_SPACE1)
        md_resource = self._create_metadef_resource(path=path,
                                                    data=data)
        self.assertEqual('MyNamespace', md_resource['namespace'])
        self.assertEqual(
            'MyResourceType',
            md_resource['resource_type_associations'][0]['name'])

        # Now disable add_metadef_resource_type_association permissions and
        # make sure that even you have permission to create namespace the
        # request will fail
        self.set_policy_rules({
            'add_metadef_resource_type_association': '!',
            'get_metadef_namespace': '@'
        })
        data.update(NAME_SPACE2)
        resp = self.api_post(path, json=data)
        self.assertEqual(403, resp.status_code)

    def test_namespace_create_with_objects(self):
        self.start_server()
        # First make sure you can create namespace and objects
        # with default policy
        path = '/v2/metadefs/namespaces'
        data = {
            "objects": [{
                "name": "MyObject",
                "description": "My object for My namespace",
                "properties": {
                    "test_property": {
                        "title": "test_property",
                        "description": "Test property for My object",
                        "type": "string"
                    },
                }
            }],
        }
        data.update(NAME_SPACE1)
        md_resource = self._create_metadef_resource(path=path,
                                                    data=data)
        self.assertEqual('MyNamespace', md_resource['namespace'])
        self.assertEqual(
            'MyObject',
            md_resource['objects'][0]['name'])

        # Now disable add_metadef_object permissions and
        # make sure that even you have permission to create namespace the
        # request will fail
        self.set_policy_rules({
            'add_metadef_object': '!',
            'get_metadef_namespace': '@'
        })
        data.update(NAME_SPACE2)
        resp = self.api_post(path, json=data)
        self.assertEqual(403, resp.status_code)

    def test_namespace_create_with_tags(self):
        self.start_server()
        # First make sure you can create namespace and tags
        # with default policy
        path = '/v2/metadefs/namespaces'
        data = {
            "tags": [{
                "name": "MyTag",
            }],
        }
        data.update(NAME_SPACE1)
        md_resource = self._create_metadef_resource(path=path,
                                                    data=data)
        self.assertEqual('MyNamespace', md_resource['namespace'])
        self.assertEqual(
            'MyTag',
            md_resource['tags'][0]['name'])

        # Now disable add_metadef_object permissions and
        # make sure that even you have permission to create namespace the
        # request will fail
        data.update(NAME_SPACE2)
        self.set_policy_rules({
            'add_metadef_tag': '!',
            'get_metadef_namespace': '@'
        })
        resp = self.api_post(path, json=data)
        self.assertEqual(403, resp.status_code)

    def test_namespace_create_with_properties(self):
        self.start_server()
        # First make sure you can create namespace and properties
        # with default policy
        path = '/v2/metadefs/namespaces'
        data = {
            "properties": {
                "TestProperty": {
                    "title": "MyTestProperty",
                    "description": "Test Property for My namespace",
                    "type": "string"
                },
            }
        }
        data.update(NAME_SPACE1)
        md_resource = self._create_metadef_resource(path=path,
                                                    data=data)
        self.assertEqual('MyNamespace', md_resource['namespace'])
        self.assertEqual(
            'MyTestProperty',
            md_resource['properties']['TestProperty']['title'])

        # Now disable add_metadef_property permissions and
        # make sure that even you have permission to create namespace the
        # request will fail
        data.update(NAME_SPACE2)
        self.set_policy_rules({
            'add_metadef_property': '!',
            'get_metadef_namespace': '@'
        })
        resp = self.api_post(path, json=data)
        self.assertEqual(403, resp.status_code)

    def test_namespace_get_basic(self):
        self.start_server()
        # First make sure create namespace works with default policy
        path = '/v2/metadefs/namespaces'
        md_resource = self._create_metadef_resource(path=path,
                                                    data=GLOBAL_NAMESPACE_DATA)
        self.assertEqual('MyNamespace', md_resource['namespace'])

        # Now make sure get_metadef_namespace will return all associated
        # resources in the response as every policy is open.
        path = "/v2/metadefs/namespaces/%s" % md_resource['namespace']
        resp = self.api_get(path)
        md_resource = resp.json
        self.assertEqual('MyNamespace', md_resource['namespace'])
        self.assertIn('objects', md_resource)
        self.assertIn('resource_type_associations', md_resource)
        self.assertIn('tags', md_resource)
        self.assertIn('properties', md_resource)

        # Now disable get_metadef_namespace policy to ensure that you are
        # forbidden to fulfill the request and get 404 not found
        self.set_policy_rules({'get_metadef_namespace': '!'})
        path = "/v2/metadefs/namespaces/%s" % md_resource['namespace']
        resp = self.api_get(path)
        self.assertEqual(404, resp.status_code)

        # Now try to get the same namespace by different user
        self.set_policy_rules({'get_metadef_namespace': '@'})
        self._verify_forbidden_converted_to_not_found(path, 'GET')

        # Now disable get_metadef_objects policy to ensure that you will
        # get forbidden response
        self.set_policy_rules({
            'get_metadef_objects': '!',
            'get_metadef_namespace': '@',
            'list_metadef_resource_types': '@',
            'get_metadef_properties': '@',
            'get_metadef_tags': '@'
        })
        path = "/v2/metadefs/namespaces/%s" % md_resource['namespace']
        resp = self.api_get(path)
        self.assertEqual(403, resp.status_code)

        # Now disable list_metadef_resource_types policy to ensure that you
        # will get forbidden response
        self.set_policy_rules({
            'get_metadef_objects': '@',
            'get_metadef_namespace': '@',
            'list_metadef_resource_types': '!',
            'get_metadef_properties': '@',
            'get_metadef_tags': '@'
        })
        path = "/v2/metadefs/namespaces/%s" % md_resource['namespace']
        resp = self.api_get(path)
        self.assertEqual(403, resp.status_code)

        # Now disable get_metadef_properties policy to ensure that you will
        # ger forbidden response
        self.set_policy_rules({
            'get_metadef_objects': '@',
            'get_metadef_namespace': '@',
            'list_metadef_resource_types': '@',
            'get_metadef_properties': '!',
            'get_metadef_tags': '@'
        })
        path = "/v2/metadefs/namespaces/%s" % md_resource['namespace']
        resp = self.api_get(path)
        self.assertEqual(403, resp.status_code)

        # Now disable get_metadef_tags policy to ensure that you will
        # get forbidden response
        self.set_policy_rules({
            'get_metadef_objects': '@',
            'get_metadef_namespace': '@',
            'list_metadef_resource_types': '@',
            'get_metadef_properties': '@',
            'get_metadef_tags': '!'
        })
        path = "/v2/metadefs/namespaces/%s" % md_resource['namespace']
        resp = self.api_get(path)
        self.assertEqual(403, resp.status_code)

    def test_namespace_update_basic(self):
        self.start_server()
        # First make sure create namespace works with default policy
        path = '/v2/metadefs/namespaces'
        md_resource = self._create_metadef_resource(path=path,
                                                    data=NAME_SPACE1)
        self.assertEqual('MyNamespace', md_resource['namespace'])
        self.assertEqual('private', md_resource['visibility'])

        # Now ensure you are able to update the namespace
        path = '/v2/metadefs/namespaces/%s' % md_resource['namespace']
        data = {
            'visibility': 'public',
            'namespace': md_resource['namespace'],
        }
        resp = self.api_put(path, json=data)
        md_resource = resp.json
        self.assertEqual('MyNamespace', md_resource['namespace'])
        self.assertEqual('public', md_resource['visibility'])

        # Now disable modify_metadef_namespace permissions and make sure
        # any other attempts results in 403 forbidden
        self.set_policy_rules({
            'modify_metadef_namespace': '!',
            'get_metadef_namespace': '@',
        })

        resp = self.api_put(path, json=data)
        self.assertEqual(403, resp.status_code)

        # Now enable modify_metadef_namespace and get_metadef_namespace
        # permissions and make sure modifying non existing results in
        # 404 NotFound
        self.set_policy_rules({
            'modify_metadef_namespace': '@',
            'get_metadef_namespace': '@',
        })
        path = '/v2/metadefs/namespaces/non-existing'
        resp = self.api_put(path, json=data)
        self.assertEqual(404, resp.status_code)

        # Note for reviewers, this causes our "check get if modify fails"
        # logic to return 404 as we expect, but not related to the latest
        # rev that checks the namespace get operation first.
        self.set_policy_rules({
            'modify_metadef_namespace': '!',
            'get_metadef_namespace': '!',
        })
        path = '/v2/metadefs/namespaces/%s' % md_resource['namespace']
        resp = self.api_put(path, json=data)
        self.assertEqual(404, resp.status_code)

        # Ensure accessing non visible namespace will catch 403 and
        # return 404 to user
        self.set_policy_rules({
            'modify_metadef_namespace': '@',
            'get_metadef_namespace': '@',
        })
        # Reset visibility to private
        # Now ensure you are able to update the namespace
        path = '/v2/metadefs/namespaces/%s' % md_resource['namespace']
        data = {
            'visibility': 'private',
            'namespace': md_resource['namespace'],
        }
        resp = self.api_put(path, json=data)
        md_resource = resp.json
        self.assertEqual('MyNamespace', md_resource['namespace'])
        self.assertEqual('private', md_resource['visibility'])

        # Now try to update the same namespace by different user
        self._verify_forbidden_converted_to_not_found(path, 'PUT',
                                                      json=data)

    def test_namespace_delete_basic(self):
        def _create_private_namespace(fn_call, data):
            path = '/v2/metadefs/namespaces'
            return fn_call(path=path, data=data)

        self.start_server()
        # First make sure create namespace works with default policy
        md_resource = _create_private_namespace(
            self._create_metadef_resource, NAME_SPACE1)
        self.assertEqual('MyNamespace', md_resource['namespace'])

        # Now ensure you are able to delete the namespace
        path = '/v2/metadefs/namespaces/%s' % md_resource['namespace']
        resp = self.api_delete(path)
        self.assertEqual(204, resp.status_code)
        # Verify that namespace is deleted
        path = "/v2/metadefs/namespaces/%s" % md_resource['namespace']
        resp = self.api_get(path)
        self.assertEqual(404, resp.status_code)

        # Now create another namespace to check deletion is not allowed
        md_resource = _create_private_namespace(
            self._create_metadef_resource, NAME_SPACE2)
        self.assertEqual('MySecondNamespace', md_resource['namespace'])

        # Now disable delete_metadef_namespace permissions and make sure
        # any other attempts fail
        path = '/v2/metadefs/namespaces/%s' % md_resource['namespace']
        self.set_policy_rules({
            'delete_metadef_namespace': '!',
            'get_metadef_namespace': '@'
        })
        resp = self.api_delete(path)
        self.assertEqual(403, resp.status_code)

        # Now enable both permissions and make sure deleting non
        # exsting namespace returns 404 NotFound
        self.set_policy_rules({
            'delete_metadef_namespace': '@',
            'get_metadef_namespace': '@'
        })
        path = '/v2/metadefs/namespaces/non-existing'
        resp = self.api_delete(path)
        self.assertEqual(404, resp.status_code)

        # Note for reviewers, this causes our "check get if delete fails"
        # logic to return 404 as we expect, but not related to the latest
        # rev that checks the namespace get operation first.
        self.set_policy_rules({
            'delete_metadef_namespace': '!',
            'get_metadef_namespace': '!',
        })
        path = '/v2/metadefs/namespaces/%s' % md_resource['namespace']
        resp = self.api_delete(path)
        self.assertEqual(404, resp.status_code)

        # Ensure accessing non visible namespace will catch 403 and
        # return 404 to user
        self.set_policy_rules({
            'delete_metadef_namespace': '@',
            'get_metadef_namespace': '@',
        })
        self._verify_forbidden_converted_to_not_found(path, 'DELETE')

    def test_namespace_delete_objects_basic(self):
        self.start_server()
        # First make sure create namespace  and object works with default
        # policy
        path = '/v2/metadefs/namespaces'
        md_resource = self._create_metadef_resource(path,
                                                    data=GLOBAL_NAMESPACE_DATA)
        self.assertEqual('MyNamespace', md_resource['namespace'])
        self.assertIn('objects', md_resource)

        # Now ensure you are able to delete the object(s) from namespace
        path = '/v2/metadefs/namespaces/%s/objects' % md_resource['namespace']
        resp = self.api_delete(path)
        self.assertEqual(204, resp.status_code)
        # Verify that object from namespace is deleted but namespace is
        # available
        path = "/v2/metadefs/namespaces/%s" % md_resource['namespace']
        resp = self.api_get(path)
        md_resource = resp.json
        self.assertNotIn('objects', md_resource)
        self.assertEqual('MyNamespace', md_resource['namespace'])

        # Now add another object to the namespace
        path = '/v2/metadefs/namespaces/%s/objects' % md_resource['namespace']
        data = {
            "name": "MyObject",
            "description": "My object for My namespace",
            "properties": {
                "test_property": {
                    "title": "test_property",
                    "description": "Test property for My object",
                    "type": "string"
                },
            }
        }
        md_object = self._create_metadef_resource(path, data=data)
        self.assertEqual('MyObject', md_object['name'])

        # Now disable delete_metadef_namespace permissions and make sure
        # any other attempts to delete objects fails
        path = '/v2/metadefs/namespaces/%s/objects' % md_resource['namespace']
        self.set_policy_rules({
            'delete_metadef_namespace': '!',
            'get_metadef_namespace': '@'
        })
        resp = self.api_delete(path)
        self.assertEqual(403, resp.status_code)

        # Now enable both permissions and make sure
        # deleting objects for non existing namespace returns 404 Not found
        path = '/v2/metadefs/namespaces/non-existing/objects'
        self.set_policy_rules({
            'delete_metadef_namespace': '@',
            'get_metaded_namespace': '@'
        })
        resp = self.api_delete(path)
        self.assertEqual(404, resp.status_code)

        # Note for reviewers, this causes our "check get if delete fails"
        # logic to return 404 as we expect, but not related to the latest
        # rev that checks the namespace get operation first.
        self.set_policy_rules({
            'delete_metadef_namespace': '!',
            'get_metadef_namespace': '!',
        })
        path = '/v2/metadefs/namespaces/%s/objects' % md_resource['namespace']
        resp = self.api_delete(path)
        self.assertEqual(404, resp.status_code)

        # Ensure accessing non visible namespace will catch 403 and
        # return 404 to user
        self.set_policy_rules({
            'delete_metadef_namespace': '@',
            'get_metadef_namespace': '@',
        })
        self._verify_forbidden_converted_to_not_found(path, 'DELETE')

    def test_namespace_delete_properties_basic(self):
        self.start_server()
        # First make sure create namespace  and properties works with default
        # policy
        path = '/v2/metadefs/namespaces'
        md_resource = self._create_metadef_resource(path,
                                                    data=GLOBAL_NAMESPACE_DATA)
        namespace = md_resource['namespace']
        self.assertEqual('MyNamespace', namespace)
        self.assertIn('properties', md_resource)

        # Now ensure you are able to delete all properties from namespace
        path = '/v2/metadefs/namespaces/%s/properties' % namespace
        resp = self.api_delete(path)
        self.assertEqual(204, resp.status_code)
        # Verify that properties from namespace are deleted but namespace is
        # available
        path = "/v2/metadefs/namespaces/%s" % namespace
        resp = self.api_get(path)
        md_resource = resp.json
        self.assertNotIn('properties', md_resource)
        self.assertEqual('MyNamespace', namespace)

        # Now add another property to the namespace
        path = '/v2/metadefs/namespaces/%s/properties' % namespace
        data = {
            "name": "MyProperty",
            "title": "test_property",
            "description": "Test property for My Namespace",
            "type": "string"
        }
        md_resource = self._create_metadef_resource(path,
                                                    data=data)
        self.assertEqual('MyProperty', md_resource['name'])

        # Now disable delete_metadef_namespace permissions and make sure
        # any other attempts to delete properties fails
        path = '/v2/metadefs/namespaces/%s/properties' % namespace
        self.set_policy_rules({
            'delete_metadef_namespace': '!',
            'get_metadef_namespace': '@',
        })
        resp = self.api_delete(path)
        self.assertEqual(403, resp.status_code)

        # Now disable both permissions and make sure
        # deleting properties for non existing namespace returns 404 Not found
        path = '/v2/metadefs/namespaces/non-existing/properties'
        self.set_policy_rules({
            'delete_metadef_namespace': '@',
            'get_metadef_namespace': '@',
        })
        resp = self.api_delete(path)
        self.assertEqual(404, resp.status_code)

        # Note for reviewers, this causes our "check get if delete fails"
        # logic to return 404 as we expect, but not related to the latest
        # rev that checks the namespace get operation first.
        self.set_policy_rules({
            'delete_metadef_namespace': '!',
            'get_metadef_namespace': '!',
        })
        path = '/v2/metadefs/namespaces/%s/properties' % namespace
        resp = self.api_delete(path)
        self.assertEqual(404, resp.status_code)

        # Ensure accessing non visible namespace will catch 403 and
        # return 404 to user
        self.set_policy_rules({
            'delete_metadef_namespace': '@',
            'get_metadef_namespace': '@',
        })
        self._verify_forbidden_converted_to_not_found(path, 'DELETE')

    def test_namespace_delete_tags_basic(self):
        self.start_server()
        # First make sure create namespace  and tags works with default
        # policy
        path = '/v2/metadefs/namespaces'
        md_resource = self._create_metadef_resource(path,
                                                    data=GLOBAL_NAMESPACE_DATA)
        namespace = md_resource['namespace']
        self.assertEqual('MyNamespace', namespace)
        self.assertIn('tags', md_resource)

        # Now ensure you are able to delete all properties from namespace
        path = '/v2/metadefs/namespaces/%s/tags' % namespace
        resp = self.api_delete(path)
        self.assertEqual(204, resp.status_code)
        # Verify that tags from namespace are deleted but namespace is
        # available
        path = "/v2/metadefs/namespaces/%s" % namespace
        resp = self.api_get(path)
        md_resource = resp.json
        self.assertNotIn('tags', md_resource)
        self.assertEqual('MyNamespace', namespace)

        # Now add another tag to the namespace
        tag_name = "MyTag"
        path = '/v2/metadefs/namespaces/%s/tags/%s' % (namespace,
                                                       tag_name)
        md_resource = self._create_metadef_resource(path)
        self.assertEqual('MyTag', md_resource['name'])

        # Now disable delete_metadef_namespace permissions and make sure
        # any other attempts to delete tags fails
        path = '/v2/metadefs/namespaces/%s/tags' % namespace
        self.set_policy_rules({
            'delete_metadef_namespace': '!',
            'get_metadef_namespace': '@'
        })
        resp = self.api_delete(path)
        self.assertEqual(403, resp.status_code)

        # Now enable delete_metadef_namespace permissions and and disable
        # delete_metadef_tags to make sure
        # any other attempts to delete tags fails
        path = '/v2/metadefs/namespaces/%s/tags' % namespace
        self.set_policy_rules({
            'delete_metadef_namespace': '@',
            'delete_metadef_tags': '!',
            'get_metadef_namespace': '@'
        })
        resp = self.api_delete(path)
        self.assertEqual(403, resp.status_code)

        # Now enable all permissions and make sure deleting tags for
        # non existing namespace will return 404 Not found
        path = '/v2/metadefs/namespaces/non-existing/tags'
        self.set_policy_rules({
            'delete_metadef_namespace': '@',
            'delete_metadef_tags': '@',
            'get_metadef_namespace': '@'
        })
        resp = self.api_delete(path)
        self.assertEqual(404, resp.status_code)

        # Note for reviewers, this causes our "check get if delete fails"
        # logic to return 404 as we expect, but not related to the latest
        # rev that checks the namespace get operation first.
        self.set_policy_rules({
            'delete_metadef_namespace': '!',
            'get_metadef_namespace': '!',
            'delete_metadef_tags': '!'
        })
        path = '/v2/metadefs/namespaces/%s/tags' % namespace
        resp = self.api_delete(path)
        self.assertEqual(404, resp.status_code)

        # Ensure accessing non visible namespace will catch 403 and
        # return 404 to user
        self.set_policy_rules({
            'delete_metadef_namespace': '@',
            'get_metadef_namespace': '@',
            'delete_metadef_tags': '@'
        })
        self._verify_forbidden_converted_to_not_found(path, 'DELETE')
